"""
T-walk self adjusting MCMC.
"""

# Author: By Andres Christen.
# see:  http://www.cimat.mx/~jac/twalk/

# 2010-04-17 Paul Kienzle
# * typo fixups
# * move pylab import to the particular functions
# * remove scipy dependence

__all__ = ["pytwalk"]

from numpy.random import uniform, normal
from numpy import ones, zeros, cumsum, shape, asmatrix as mat, cov, mean, ceil, matrix, sqrt
from numpy import floor, exp, log, sum, pi, arange

# Some auxiliary functions and constants
# square of the norm.


def SqrNorm(x):
    return sum(x * x)


log2pi = log(2 * pi)


class pytwalk:
    """This is the t-walk class.

    Initiates defining the dimension= n and -log of the objective function= U,
    Supp defines the support, returns True if x within the support, eg:

    Mytwalk = twalk( n=3, U=MyMinusLogf, Supp=MySupportFunction).

    Then do: Mytwalk.Run?

    Other parameter are:
    ww= the prob. of choosing each kernel, aw, at, n1phi (see inside twalk.py)
    with default values as in the paper, normally NOT needed to be changed."""

    def __init__(
        self,
        n,
        U=(lambda x: sum(0.5 * x**2)),
        Supp=(lambda x: True),
        ww=(0.0000, 0.4918, 0.4918, 0.0082, 0.0082),
        aw=1.5,
        at=6.0,
        n1phi=4.0,
    ):
        self.n = n
        self.U = U
        self.Supp = Supp
        self.Output = zeros((1, n + 1))  # No data (MCMC output) yet
        self.T = 1
        # To save the acceptance rates of each kernel, and the global acc. rate
        self.Acc = zeros(6)

        # Kernel probabilities
        self.Fw = cumsum(ww)

        # Parameters for the propolsals
        self.aw = aw  # For the walk move
        self.at = at  # For the Traverse move

        # n1phi = 5 ### expected value of parameters to move
        self.pphi = min(n, n1phi) / (1.0 * n)  # Prob. of choosing each par.

    def Run(self, T, x0, xp0):
        """Run the twalk.

        Run( T, x0, xp0),
        T = Number of iterations.
        x0, xp0, two initial points within the support,
        ***each entry of x0 and xp0 most be different***.
        """

        print("twalk: Running the twalk with %d iterations." % T)
        # Check x0 and xp0 in the support
        x = x0  # Reference, so we can retrieve the last values used
        if not (self.Supp(x)):
            print("twalk: ERROR, initial point x0 = %12.4g out of support." % x0)
            return 0
        u = self.U(x)

        xp = xp0
        if not (self.Supp(xp)):
            print("twalk: ERROR, initial point xp0 = %12.4g out of support." % xp0)
            return 0
        up = self.U(xp)

        if any(abs(x0 - xp0) <= 0):
            print("twalk: ERROR, not all entries of initial values different.")
            return 0

        # Set the array to place the iterations and the U's ... we donot save
        # up's
        self.Output = zeros((T + 1, self.n + 1))
        self.T = T + 1
        self.Acc = zeros(6)
        kercall = zeros(6)  # Times each kernel is called

        # Make local references for less writing
        n = self.n
        Output = self.Output
        Acc = self.Acc

        Output[0, 0:n] = x.copy()
        Output[0, n] = u

        # Sampling
        for it in range(T):
            y, yp, ke, A, u_prop, up_prop = self.onemove(x, u, xp, up)

            kercall[ke] += 1
            kercall[5] += 1
            if uniform() < A:
                x = y.copy()  # Accept the proposal y
                u = u_prop
                xp = yp.copy()  # Accept the proposal yp
                up = up_prop

                Acc[ke] += 1
                Acc[5] += 1

            # To retrive the current values
            self.x = x
            self.xp = xp
            self.u = u
            self.up = up

            Output[it + 1, 0:n] = x.copy()
            Output[it + 1, n] = u

        if Acc[5] == 0:
            print("twalk: WARNING,  all propolsals were rejected!")
            return 0

        for i in range(6):
            if kercall[i] != 0:
                Acc[i] /= kercall[i]
        return 1

    def onemove(self, x, u, xp, up):
        """One move of the twalk.  This is basically the raw twalk kernel.
        It is usefull if the twalk is needed inside a more complex MCMC.

        onemove(x, u, xp, up),
        x, xp, two points WITHIN the support ***each entry of x0 and xp0 must be different***.
        and the value of the objective at x, and xp
        u=U(x), up=U(xp).

        It returns: [y, yp, ke, A, u_prop, up_prop]
        y, yp: the proposed jump
        ke: The kernel used, 0=nothing, 1=Walk, 2=Traverse, 3=Blow, 4=Hop
        A: the M-H ratio
        u_prop, up_prop: The values for the objective func. at the proposed jumps
        """

        # Make local references for less writing
        U = self.U
        Supp = self.Supp
        Fw = self.Fw

        ker = uniform()  # To choose the kernel to be used
        ke = 1
        A = 0

        # Kernel nothing exchange x with xp
        if (0.0 <= ker) & (ker < Fw[0]):
            ke = 0
            y = xp.copy()
            up_prop = u
            yp = x.copy()
            u_prop = up
            # A is the MH acceptance ratio
            A = 1.0
            # always accepted

        # The Walk move
        if (Fw[0] <= ker) & (ker < Fw[1]):
            ke = 1

            dir = uniform()

            if (0 <= dir) & (dir < 0.5):  # x as pivot
                yp = self.SimWalk(xp, x)

                y = x.copy()
                u_prop = u

                if (Supp(yp)) & (all(abs(yp - y) > 0)):
                    up_prop = U(yp)
                    A = exp(up - up_prop)
                else:
                    up_prop = None
                    A = 0
                    # out of support, not accepted

            else:  # xp as pivot
                y = self.SimWalk(x, xp)

                yp = xp.copy()
                up_prop = up

                if (Supp(y)) & (all(abs(yp - y) > 0)):
                    u_prop = U(y)
                    A = exp(u - u_prop)
                else:
                    u_prop = None
                    A = 0
                    # out of support, not accepted

        # The Traverse move
        if (Fw[1] <= ker) & (ker < Fw[2]):
            ke = 2
            dir = uniform()

            if (0 <= dir) & (dir < 0.5):  # x as pivot
                beta = self.Simbeta()
                yp = self.SimTraverse(xp, x, beta)

                y = x.copy()
                u_prop = u

                if Supp(yp):
                    up_prop = U(yp)
                    if self.nphi == 0:
                        A = 1  # Nothing moved
                    else:
                        A = exp((up - up_prop) + (self.nphi - 2) * log(beta))
                else:
                    up_prop = None
                    A = 0  # out of support, not accepted
            else:  # xp as pivot
                beta = self.Simbeta()
                y = self.SimTraverse(x, xp, beta)

                yp = xp.copy()
                up_prop = up

                if Supp(y):
                    u_prop = U(y)
                    if self.nphi == 0:
                        A = 1  # Nothing moved
                    else:
                        A = exp((u - u_prop) + (self.nphi - 2) * log(beta))
                else:
                    u_prop = None
                    A = 0  # out of support, not accepted

        # The Blow move
        if (Fw[2] <= ker) & (ker < Fw[3]):
            ke = 3
            dir = uniform()

            if (0 <= dir) & (dir < 0.5):  # x as pivot
                yp = self.SimBlow(xp, x)

                y = x.copy()
                u_prop = u
                if (Supp(yp)) & all(yp != x):
                    up_prop = U(yp)
                    W1 = self.GBlowU(yp, xp, x)
                    W2 = self.GBlowU(xp, yp, x)
                    A = exp((up - up_prop) + (W1 - W2))
                else:
                    up_prop = None
                    A = 0  # out of support, not accepted
            else:  # xp as pivot
                y = self.SimBlow(x, xp)

                yp = xp.copy()
                up_prop = up
                if (Supp(y)) & all(y != xp):
                    u_prop = U(y)
                    W1 = self.GBlowU(y, x, xp)
                    W2 = self.GBlowU(x, y, xp)
                    A = exp((u - u_prop) + (W1 - W2))
                else:
                    u_prop = None
                    A = 0  # out of support, not accepted

        # The Hop move
        if (Fw[3] <= ker) & (ker < Fw[4]):
            ke = 4
            dir = uniform()

            if (0 <= dir) & (dir < 0.5):  # x as pivot
                yp = self.SimHop(xp, x)

                y = x.copy()
                u_prop = u
                if (Supp(yp)) & all(yp != x):
                    up_prop = U(yp)
                    W1 = self.GHopU(yp, xp, x)
                    W2 = self.GHopU(xp, yp, x)
                    A = exp((up - up_prop) + (W1 - W2))
                else:
                    up_prop = None
                    A = 0  # out of support, not accepted
            else:  # xp as pivot
                y = self.SimHop(x, xp)

                yp = xp.copy()
                up_prop = up
                if (Supp(y)) & all(y != xp):
                    u_prop = U(y)
                    W1 = self.GHopU(y, x, xp)
                    W2 = self.GHopU(x, y, xp)
                    A = exp((u - u_prop) + (W1 - W2))
                else:
                    u_prop = None
                    A = 0  # out of support, not accepted

        return [y, yp, ke, A, u_prop, up_prop]

    ##########################################################################
    # Auxiliaries for the kernels

    # Used by the Walk kernel
    def SimWalk(self, x, xp):
        aw = self.aw
        n = self.n

        phi = uniform(size=n) < self.pphi  # parameters to move
        self.nphi = sum(phi)
        z = zeros(n)

        for i in range(n):
            if phi[i]:
                u = uniform()
                z[i] = (aw / (1 + aw)) * (aw * u**2.0 + 2.0 * u - 1.0)

        return x + (x - xp) * z

    # Used by the Traverse kernel
    def Simbeta(self):
        at = self.at
        if uniform() < (at - 1.0) / (2.0 * at):
            return exp(1.0 / (at + 1.0) * log(uniform()))
        else:
            return exp(1.0 / (1.0 - at) * log(uniform()))

    def SimTraverse(self, x, xp, beta):
        n = self.n

        phi = uniform(size=n) < self.pphi
        self.nphi = sum(phi)

        rt = x.copy()
        for i in range(n):
            if phi[i]:
                rt[i] = xp[i] + beta * (xp[i] - x[i])

        return rt

    # Used by the Blow kernel
    def SimBlow(self, x, xp):
        n = self.n

        phi = uniform(size=n) < self.pphi
        self.nphi = sum(phi)

        self.sigma = max(phi * abs(xp - x))

        rt = x.copy()
        for i in range(n):
            if phi[i]:
                rt[i] = x[i] + self.sigma * normal()

        return rt

    def GBlowU(self, h, x, xp):
        nphi = self.nphi

        if nphi > 0:
            return (nphi / 2.0) * log2pi + nphi * log(self.sigma) + 0.5 * SqrNorm(h - xp) / (self.sigma**2)
        else:
            return 0

    # Used by the Hop kernel
    def SimHop(self, x, xp):
        n = self.n

        phi = uniform(size=n) < self.pphi
        self.nphi = sum(phi)

        self.sigma = max(phi * abs(xp - x)) / 3.0

        rt = x.copy()
        for i in range(n):
            if phi[i]:
                rt[i] = xp[i] + self.sigma * normal()

        return rt

    def GHopU(self, h, x, xp):  # It is actually equal to GBlowU!
        nphi = self.nphi

        if nphi > 0:
            return (nphi / 2.0) * log2pi + nphi * log(self.sigma) + 0.5 * SqrNorm(h - xp) / (self.sigma**2)
        else:
            return 0

    ##########################################################################
    # Output analysis auxiliary methods

    def IAT(self, par=-1, start=0, end=0, maxlag=0):
        """Calculate the Integrated Autocorrelation Times of parameters par
        the default value par=-1 is for the IAT of the U's"""
        if end == 0:
            end = self.T

        if self.Acc[5] == 0:
            print("twalk: IAT: WARNING,  all proposals were rejected!")
            print("twalk: IAT: Cannot calculate IAT, fixing it to the sample size.")
            return self.T

        iat = IAT(self.Output, cols=par, maxlag=maxlag, start=start, end=end)

        return iat

    def TS(self, par=-1, start=0, end=0):
        """Plot time series of parameter par (default = log f) etc."""
        from pylab import plot, xlabel, ylabel

        if par == -1:
            par = self.n

        if end == 0:
            end = self.T

        if par == self.n:
            plot(arange(start, end), -1 * self.Output[start:end, par])
            ylabel("Log of Objective")
        else:
            plot(arange(start, end), self.Output[start:end, par])
            ylabel("Parameter %d" % par)
        xlabel("Iteration")

    def Ana(self, par=-1, start=0, end=0):
        """Output Analysis, TS plots, acceptance rates, IAT etc."""
        if par == -1:
            par = self.n

        if end == 0:
            end = self.T

        print("Acceptance rates for the Walk, Traverse, Blow and Hop kernels:" + str(self.Acc[1:5]))
        print("Global acceptance rate: %7.5f" % self.Acc[5])

        iat = self.IAT(par=par, start=start, end=end)
        print("Integrated Autocorrelation Time: %7.1f, IAT/n: %7.1f" % (iat, iat / self.n))

        self.TS(par=par, start=start, end=end)

    def Hist(self, par=-1, start=0, end=0, g=(lambda x: x[0]), xlab="g", bins=20):
        """Basic histograms and output analysis.  If par=-1, use g.
        The function g provides a transformation to be applied to the data,
        eg g=(lambda x: abs(x[0]-x[1]) would plot a histogram of the distance
        between parameters 0 and 1, etc."""
        from pylab import hist, xlabel

        if end == 0:
            end = self.T

        if par == -1:
            ser = zeros(end - start)
            for it in range(end - start):
                ser[it] = g(self.Output[it + start, :])
            xlabel(xlab)
            print("Mean for %s= %f" % (xlab, mean(ser)))
        else:
            ser = self.Output[start:end, par]
            xlabel("Parameter %d" % par)
            print("Mean for par %d= %f" % (par, mean(ser)))

        hist(ser, bins=bins)
        print("Do:\nfrom pylab import show\nshow()")

    def Save(self, fnam, start=0, thin=1):
        """Saves the Output as a text file, starting at start (burn in), with thinning (thin)."""
        print(("Saving output, all pars. plus the U's in file", fnam))

        from numpy import savetxt

        savetxt(fnam, self.Output[start:,])

    # A simple Random Walk M-H
    def RunRWMH(self, T, x0, sigma):
        """Run a simple Random Walk M-H"""

        print("twalk: This is the Random Walk M-H running with %d iterations." % T)
        # Local variables
        x = x0.copy()
        if not (self.Supp(x)):
            print("twalk: ERROR, initial point x0 out of support.")
            return 0

        u = self.U(x)
        n = self.n

        # Set the array to place the iterations and the U's
        self.Output = zeros((T + 1, n + 1))
        self.Acc = zeros(6)

        # Make local references for less writing
        Output = self.Output
        U = self.U
        Supp = self.Supp
        Acc = self.Acc

        Output[0, 0:n] = x.copy()
        Output[0, n] = u

        y = x.copy()
        for it in range(T):
            y = x + normal(size=n) * sigma  # each entry with sigma[i] variance
            if Supp(y):  # If it is within the support of the objective
                uprop = U(y)  # Evaluate the objective
                if uniform() < exp(u - uprop):
                    x = y.copy()  # Accept the propolsal y
                    u = uprop
                    Acc[5] += 1

            Output[it + 1, 0:n] = x
            Output[it + 1, n] = u

        if Acc[5] == 0:
            print("twalk: WARNING,  all propolsals were rejected!")
            return 0

        Acc[5] /= T
        return 1


##########################################################################
# Auxiliary functions to calculate Integrated autocorrelation times of a
# time series ####


# Calculates an autocovariance 2x2 matrix at lag l in column c of matrix Ser with T rows
# The variances of each series are in the diagonal and the
# (auto)covariance in the off diag.
def AutoCov(Ser, c, la, T=0):
    if T == 0:
        T = shape(Ser)[0]  # Number of rows in the matrix (sample size)

    return cov(Ser[0 : (T - 1 - la), c], Ser[la : (T - 1), c], bias=1)


# Calculates the autocorrelation from lag 0 to lag la of columns cols (list)
# for matrix Ser
def AutoCorr(Ser, cols=0, la=1):
    T = shape(Ser)[0]  # Number of rows in the matrix (sample size)

    ncols = shape(mat(cols))[1]  # Number of columns to analyse (parameters)

    # if ncols == 1:
    #    cols = [cols]

    # Matrix to hold output
    Out = matrix(ones((la) * ncols)).reshape(la, ncols)

    for c in range(ncols):
        for l in range(1, la + 1):
            Co = AutoCov(Ser, cols[c], l, T)
            Out[l - 1, c] = Co[0, 1] / (sqrt(Co[0, 0] * Co[1, 1]))

    return Out


# Makes an upper band matrix of ones, to add the autocorrelation matrix
# gamma = auto[2*m+1,c]+auto[2*m+2,c] etc.
# MakeSumMat(lag) * AutoCorr( Ser, cols=c, la=lag) to make the gamma matrix
def MakeSumMat(lag):
    rows = (lag) / 2  # Integer division!
    Out = mat(zeros([rows, lag], dtype=int))

    for i in range(rows):
        Out[i, 2 * i] = 1
        Out[i, 2 * i + 1] = 1

    return Out


# Finds the cutting time, when the gammas become negative
def Cutts(Gamma):
    cols = shape(Gamma)[1]
    rows = shape(Gamma)[0]
    Out = mat(zeros([1, cols], dtype=int))
    Stop = mat(zeros([1, cols], dtype=bool))

    if rows == 1:
        return Out

    i = 0
    # while (not(all(Stop)) & (i < (rows-1))):
    for i in range(rows - 1):
        for j in range(cols):  # while Gamma stays positive and decreasing
            if ((Gamma[i + 1, j] > 0.0) & (Gamma[i + 1, j] < Gamma[i, j])) & (not Stop[0, j]):
                Out[0, j] = i + 1  # the cutting time for colomn j is i+i
            else:
                Stop[0, j] = True
        i += 1

    return Out


# Automatically find a maxlag for IAT calculations
def AutoMaxlag(Ser, c, rholimit=0.05, maxmaxlag=20000):
    Co = AutoCov(Ser, c, la=1)
    rho = Co[0, 1] / Co[0, 0]  # lag one autocorrelation

    # if autocorrelation is like exp(- lag/lam) then, for lag = 1
    lam = -1.0 / log(abs(rho))

    # Our initial guess for maxlag is 1.5 times lam (eg. three times the mean
    # life)
    maxlag = int(floor(3.0 * lam)) + 1

    # We take 1% of lam to jump forward and look for the
    # rholimit threshold
    jmp = int(ceil(0.01 * lam)) + 1

    T = shape(Ser)[0]  # Number of rows in the matrix (sample size)

    while (abs(rho) > rholimit) & (maxlag < min(T / 2, maxmaxlag)):
        Co = AutoCov(Ser, c, la=maxlag)
        rho = Co[0, 1] / Co[0, 0]
        maxlag = maxlag + jmp
        ###print("maxlag=", maxlag, "rho", abs(rho), "\n")

    maxlag = int(floor(1.3 * maxlag))
    # 30% more

    if maxlag >= min(T / 2, maxmaxlag):  # not enough data
        fixmaxlag = min(min(T / 2, maxlag), maxmaxlag)
        print(
            "AutoMaxlag: Warning: maxlag= %d > min(T/2,maxmaxlag=%d), fixing it to %d" % (maxlag, maxmaxlag, fixmaxlag)
        )
        return fixmaxlag

    if maxlag <= 1:
        fixmaxlag = 10
        print("AutoMaxlag: Warning: maxlag= %d ?!, fixing it to %d" % (maxlag, fixmaxlag))
        return fixmaxlag

    print("AutoMaxlag: maxlag= %d." % maxlag)
    return maxlag


# Find the IAT
def IAT(Ser, cols=-1, maxlag=0, start=0, end=0):
    ncols = shape(mat(cols))[1]  # Number of columns to analyse (parameters)
    if ncols == 1:
        if cols == -1:
            cols = shape(Ser)[1] - 1  # default = last column
        cols = [cols]

    if end == 0:
        end = shape(Ser)[0]

    if maxlag == 0:
        for c in cols:
            maxlag = max(maxlag, AutoMaxlag(Ser[start:end, :], c))

    # print("IAT: Maxlag=", maxlag)

    # Ga = MakeSumMat(maxlag) * AutoCorr( Ser[start:end,:], cols=cols, la=maxlag)

    Ga = mat(zeros((maxlag / 2, ncols)))
    auto = AutoCorr(Ser[start:end, :], cols=cols, la=maxlag)

    # Instead of producing the maxlag/2 X maxlag MakeSumMat matrix, we
    # calculate the gammas like this
    for c in range(ncols):
        for i in range(maxlag / 2):
            Ga[i, c] = auto[2 * i, c] + auto[2 * i + 1, c]

    cut = Cutts(Ga)
    nrows = shape(Ga)[0]

    ncols = shape(cut)[1]
    Out = -1.0 * mat(ones([1, ncols]))

    if any((cut + 1) == nrows):
        print("IAT: Warning: Not enough lag to calculate IAT")

    for c in range(ncols):
        for i in range(cut[0, c] + 1):
            Out[0, c] += 2 * Ga[i, c]

    return Out



# This program is public domain
"""
BSpline calculator.

Given a set of knots, compute the cubic B-spline interpolation.
"""

__all__ = ["bspline", "pbs"]

import numpy as np
from numpy import maximum as max, minimum as min


def pbs(x, y, t, clamp=True, parametric=True):
    """
    Evaluate the parametric B-spline px(t),py(t).

    *x* and *y* are the control points, and *t* are the points
    in [0,1] at which they are evaluated.   The *x* values are
    sorted so that the spline describes a function.

    The spline goes through the control points at the ends. If *clamp*
    is True, the derivative of the spline at both ends is zero. If *clamp*
    is False, the derivative at the ends is equal to the slope connecting
    the final pair of control points.

    If *parametric* is False, then parametric points t' are chosen such
    that x(t') = *t*.

    The B-spline knots are chosen to be equally spaced within [0,1].
    """
    x = list(sorted(x))
    knot = np.hstack((0, 0, np.linspace(0, 1, len(y)), 1, 1))
    cx = np.hstack((x[0], x[0], x[0], (2 * x[0] + x[1]) / 3, x[1:-1], (2 * x[-1] + x[-2]) / 3, x[-1]))
    if clamp:
        cy = np.hstack((y[0], y[0], y[0], y, y[-1]))
    else:
        cy = np.hstack((y[0], y[0], y[0], y[0] + (y[1] - y[0]) / 3, y[1:-1], y[-1] + (y[-2] - y[-1]) / 3, y[-1]))

    if parametric:
        return _bspline3(knot, cx, t), _bspline3(knot, cy, t)

    # Find parametric t values corresponding to given z values
    # First try a few newton steps
    xt = np.interp(t, x, np.linspace(0, 1, len(x)))
    with np.errstate(all="ignore"):
        for _ in range(6):
            pt, dpt = _bspline3(knot, cx, xt, nderiv=1)
            xt -= (pt - t) / dpt
        idx = np.isnan(xt) | (abs(_bspline3(knot, cx, xt) - t) > 1e-9)

    # Use bisection when newton fails
    if idx.any():
        missing = t[idx]
        # print missing
        t_lo, t_hi = 0 * missing, 1 * missing
        for _ in range(30):  # bisection with about 1e-9 tolerance
            trial = (t_lo + t_hi) / 2
            ptrial = _bspline3(knot, cx, trial)
            tidx = ptrial < missing
            t_lo[tidx] = trial[tidx]
            t_hi[~tidx] = trial[~tidx]
        xt[idx] = (t_lo + t_hi) / 2
    # print "err",np.max(abs(_bspline3(knot,cx,t)-xt))

    # Return y evaluated at the interpolation points
    return _bspline3(knot, cx, xt), _bspline3(knot, cy, xt)


def bspline(y, xt, clamp=True):
    """
    Evaluate the B-spline with control points *y* at positions *xt* in [0,1].

    The spline goes through the control points at the ends.  If *clamp*
    is True, the derivative of the spline at both ends is zero.  If *clamp*
    is False, the derivative at the ends is equal to the slope connecting
    the final pair of control points.

    B-spline knots are chosen to be equally spaced within [0,1].
    """
    knot = np.hstack((0, 0, np.linspace(0, 1, len(y)), 1, 1))
    if clamp:
        cy = np.hstack(([y[0]] * 3, y, y[-1]))
    else:
        cy = np.hstack((y[0], y[0], y[0], y[0] + (y[1] - y[0]) / 3, y[1:-1], y[-1] + (y[-2] - y[-1]) / 3, y[-1]))
    return _bspline3(knot, cy, xt)


def _bspline3(knot, control, t, nderiv=0):
    """
    Evaluate the B-spline specified by the given *knot* sequence and
    *control* values at the parametric points *t*.  *nderiv* selects
    the function or derivative to evaluate.
    """
    knot, control, t = [np.asarray(v) for v in (knot, control, t)]

    # Deal with values outside the range
    valid = (t > knot[0]) & (t <= knot[-1])
    tv = t[valid]
    f = np.zeros(t.shape)
    f[t <= knot[0]] = control[0]
    f[t >= knot[-1]] = control[-1]

    # Find B-Spline parameters for the individual segments
    end = len(knot) - 1
    segment = knot.searchsorted(tv) - 1
    tm2 = knot[max(segment - 2, 0)]
    tm1 = knot[max(segment - 1, 0)]
    tm0 = knot[max(segment - 0, 0)]
    tp1 = knot[min(segment + 1, end)]
    tp2 = knot[min(segment + 2, end)]
    tp3 = knot[min(segment + 3, end)]

    p4 = control[min(segment + 3, end)]
    p3 = control[min(segment + 2, end)]
    p2 = control[min(segment + 1, end)]
    p1 = control[min(segment + 0, end)]

    # Compute second and third derivatives.
    if nderiv > 1:
        # Normally we require a recursion for Q, R and S to compute
        # df, d2f and d3f respectively, however Q can be computed directly
        # from intermediate values of P, S has a recursion of depth 0,
        # which leaves only the R recursion of depth 1 in the calculation
        # below.
        q4 = (p4 - p3) * 3 / (tp3 - tm0)
        q3 = (p3 - p2) * 3 / (tp2 - tm1)
        q2 = (p2 - p1) * 3 / (tp1 - tm2)
        r4 = (q4 - q3) * 2 / (tp2 - tm0)
        r3 = (q3 - q2) * 2 / (tp1 - tm1)
        if nderiv > 2:
            s4 = (r4 - r3) / (tp1 - tm0)
            d3f = np.zeros(t.shape)
            d3f[valid] = s4
        r4 = ((tv - tm0) * r4 + (tp1 - tv) * r3) / (tp1 - tm0)
        d2f = np.zeros(t.shape)
        d2f[valid] = r4

    # Compute function value and first derivative
    p4 = ((tv - tm0) * p4 + (tp3 - tv) * p3) / (tp3 - tm0)
    p3 = ((tv - tm1) * p3 + (tp2 - tv) * p2) / (tp2 - tm1)
    p2 = ((tv - tm2) * p2 + (tp1 - tv) * p1) / (tp1 - tm2)
    p4 = ((tv - tm0) * p4 + (tp2 - tv) * p3) / (tp2 - tm0)
    p3 = ((tv - tm1) * p3 + (tp1 - tv) * p2) / (tp1 - tm1)
    if nderiv >= 1:
        df = np.zeros(t.shape)
        df[valid] = (p4 - p3) * 3 / (tp1 - tm0)
    p4 = ((tv - tm0) * p4 + (tp1 - tv) * p3) / (tp1 - tm0)
    f[valid] = p4

    if nderiv == 0:
        return f
    elif nderiv == 1:
        return f, df
    elif nderiv == 2:
        return f, df, d2f
    else:
        return f, df, d2f, d3f


"""
def bspline_control(y, clamp=True):
    return _find_control(y, clamp=clamp)


def pbs_control(x, y, clamp=True):
    return _find_control(x, clamp=clamp), _find_control(y, clamp=clamp)


def _find_control(v, clamp=True):
    raise NotImplementedError("B-spline interpolation doesn't work yet")
    from scipy.linalg import solve_banded
    n = len(v)
    udiag = np.hstack([0, 0, 0, [1 / 6] * (n - 3), 0.25, 0.3])
    ldiag = np.hstack([-0.3, 0.25, [1 / 6] * (n - 3), 0, 0, 0])
    mdiag = np.hstack([1, 0.3, 7 / 12, [2 / 3] * (n - 4), 7 / 12, -0.3, 1])
    A = np.vstack([ldiag, mdiag, udiag])
    if clamp:
        # First derivative is zero at ends
        bl, br = 0, 0
    else:
        # First derivative at ends follows line between final control points
        bl, br = (v[1] - v[0]) * n, (v[-1] - v[-2]) * n
    b = np.hstack([v[0], bl, v[1:n - 1], br, v[-1]])
    x = solve_banded((1, 1), A, b)
    return x  # x[1:-1]
"""

# ===========================================================================
# test code


def speed_check():
    """
    Print the time to evaluate 400 points on a 7 knot spline.
    """
    import time

    x = np.linspace(0, 1, 7)
    x[1], x[-2] = x[2], x[-3]
    y = [9, 11, 2, 3, 8, 0, 2]
    t = np.linspace(0, 1, 400)
    t0 = time.time()
    for _ in range(1000):
        bspline(y, t, clamp=True)
    print("bspline (ms)", (time.time() - t0) / 1000)


def _check(expected, got, tol):
    """
    Check that value matches expected within tolerance.

    If *expected* is never zero, use relative error for tolerance.
    """
    relative = (np.isscalar(expected) and expected != 0) or (not np.isscalar(expected) and all(expected != 0))
    if relative:
        norm = np.linalg.norm((expected - got) / expected)
    else:
        norm = np.linalg.norm(expected - got)
    if norm >= tol:
        msg = [
            "expected %s" % str(expected),
            "got %s" % str(got),
            "tol %s norm %s" % (tol, norm),
        ]
        raise ValueError("\n".join(msg))


def _derivs(x, y):
    """
    Compute numerical derivative for a function evaluated on a fine grid.
    """
    # difference formula
    return (y[1] - y[0]) / (x[1] - x[0]), (y[-1] - y[-2]) / (x[-1] - x[-2])
    # 5-point difference formula
    # left = (y[0]-8*y[1]+8*y[3]-y[4]) / 12 / (x[1]-x[0])
    # right = (y[-5]-8*y[-4]+8*y[-2]-y[-1]) / 12 / (x[-1]-x[-2])
    # return left, right


def test():
    """bspline tests"""
    h = 1e-10
    t = np.linspace(0, 1, 100)
    dt = np.array([0, h, 2 * h, 3 * h, 4 * h, 1 - 4 * h, 1 - 3 * h, 1 - 2 * h, 1 - h, 1])
    y = [9, 11, 2, 3, 8, 0, 2]
    n = len(y)
    xeq = np.linspace(0, 1, n)
    x = xeq + 0
    x[0], x[-1] = (x[0] + x[1]) / 2, (x[-2] + x[-1]) / 2
    dx = np.array(
        [
            x[0],
            x[0] + h,
            x[0] + 2 * h,
            x[0] + 3 * h,
            x[0] + 4 * h,
            x[-1] - 4 * h,
            x[-1] - 3 * h,
            x[-1] - 2 * h,
            x[-1] - h,
            x[-1],
        ]
    )

    # ==== Check that bspline mat:w
    # ches pbs with equally spaced x

    yt = bspline(y, t, clamp=True)
    xtp, ytp = pbs(xeq, y, t, clamp=True, parametric=False)
    _check(t, xtp, 1e-8)
    _check(yt, ytp, 1e-8)

    xtp, ytp = pbs(xeq, y, t, clamp=True, parametric=True)
    _check(t, xtp, 1e-8)
    _check(yt, ytp, 1e-8)

    yt = bspline(y, t, clamp=False)
    xtp, ytp = pbs(xeq, y, t, clamp=False, parametric=False)
    _check(t, xtp, 1e-8)
    _check(yt, ytp, 1e-8)

    xtp, ytp = pbs(xeq, y, t, clamp=False, parametric=True)
    _check(t, xtp, 1e-8)
    _check(yt, ytp, 1e-8)

    # ==== Check bspline f at end points

    yt = bspline(y, t, clamp=True)
    _check(y[0], yt[0], 1e-12)
    _check(y[-1], yt[-1], 1e-12)

    yt = bspline(y, t, clamp=False)
    _check(y[0], yt[0], 1e-12)
    _check(y[-1], yt[-1], 1e-12)

    xt, yt = pbs(x, y, t, clamp=True, parametric=False)
    _check(x[0], xt[0], 1e-8)
    _check(x[-1], xt[-1], 1e-8)
    _check(y[0], yt[0], 1e-8)
    _check(y[-1], yt[-1], 1e-8)

    xt, yt = pbs(x, y, t, clamp=True, parametric=True)
    _check(x[0], xt[0], 1e-8)
    _check(x[-1], xt[-1], 1e-8)
    _check(y[0], yt[0], 1e-8)
    _check(y[-1], yt[-1], 1e-8)

    xt, yt = pbs(x, y, t, clamp=False, parametric=False)
    _check(x[0], xt[0], 1e-8)
    _check(x[-1], xt[-1], 1e-8)
    _check(y[0], yt[0], 1e-8)
    _check(y[-1], yt[-1], 1e-8)

    xt, yt = pbs(x, y, t, clamp=False, parametric=True)
    _check(x[0], xt[0], 1e-8)
    _check(x[-1], xt[-1], 1e-8)
    _check(y[0], yt[0], 1e-8)
    _check(y[-1], yt[-1], 1e-8)

    # ==== Check f' at end points
    yt = bspline(y, dt, clamp=True)
    left, right = _derivs(dt, yt)
    _check(0, left, 1e-8)
    _check(0, right, 1e-8)

    xt, yt = pbs(x, y, dx, clamp=True, parametric=False)
    left, right = _derivs(xt, yt)
    _check(0, left, 1e-8)
    _check(0, right, 1e-8)

    xt, yt = pbs(x, y, dt, clamp=True, parametric=True)
    left, right = _derivs(xt, yt)
    _check(0, left, 1e-8)
    _check(0, right, 1e-8)

    yt = bspline(y, dt, clamp=False)
    left, right = _derivs(dt, yt)
    _check((y[1] - y[0]) * (n - 1), left, 5e-4)
    _check((y[-1] - y[-2]) * (n - 1), right, 5e-4)

    xt, yt = pbs(x, y, dx, clamp=False, parametric=False)
    left, right = _derivs(xt, yt)
    _check((y[1] - y[0]) / (x[1] - x[0]), left, 5e-4)
    _check((y[-1] - y[-2]) / (x[-1] - x[-2]), right, 5e-4)

    xt, yt = pbs(x, y, dt, clamp=False, parametric=True)
    left, right = _derivs(xt, yt)
    _check((y[1] - y[0]) / (x[1] - x[0]), left, 5e-4)
    _check((y[-1] - y[-2]) / (x[-1] - x[-2]), right, 5e-4)

    # ==== Check interpolator
    # yc = bspline_control(y)
    # print("y",y)
    # print("p(yc)",bspline(yc,xeq))


def demo():
    """Show bsline curve for a set of control points."""
    from pylab import linspace, subplot, plot, legend, show

    # y = [9 ,6, 1, 3, 8, 4, 2]
    # y = [9, 11, 13, 3, -2, 0, 2]
    y = [9, 11, 2, 3, 8, 0]
    # y = [9, 9, 1, 3, 8, 2, 2]
    x = linspace(0, 1, len(y))
    t = linspace(x[0], x[-1], 400)
    subplot(211)
    plot(t, bspline(y, t, clamp=False), "-.y", label="unclamped bspline")  # bspline
    # bspline
    plot(t, bspline(y, t, clamp=True), "-y", label="clamped bspline")
    plot(sorted(x), y, ":oy", label="control points")
    legend()
    # left, right = _derivs(t, bspline(y, t, clamp=False))
    # print(left, (y[1] - y[0]) / (x[1] - x[0]))

    subplot(212)
    xt, yt = pbs(x, y, t, clamp=False)
    plot(xt, yt, "-.b", label="unclamped pbs")  # pbs
    xt, yt = pbs(x, y, t, clamp=True)
    plot(xt, yt, "-b", label="clamped pbs")  # pbs
    # xt,yt = pbs(x,y,t,clamp=True, parametric=True)
    # plot(xt,yt,'-g') # pbs
    plot(sorted(x), y, ":ob", label="control points")
    legend()
    show()


# B-Spline control point inverse function is not yet implemented
"""
def demo_interp():
    from pylab import linspace, plot, show
    x = linspace(0, 1, 7)
    y = [9, 11, 2, 3, 8, 0, 2]
    t = linspace(0, 1, 400)
    yc = bspline_control(y, clamp=True)
    xc = linspace(x[0], x[-1], 9)
    plot(xc, yc, ':oy', x, y, 'xg')
    #knot = np.hstack((0, np.linspace(0,1,len(y)), 1))
    #fy = _bspline3(knot,yc,t)
    fy = bspline(yc, t, clamp=True)
    plot(t, fy, '-.y')
    show()
"""

if __name__ == "__main__":
    # test()
    demo()
    # demo_interp()
    # speed_check()



"""
Package data files
==================

Some python packages, particularly gui packages, but also other packages
with static data need to be able to ship the data with the package.  This
is particularly a problem for py2exe since python does not provide any
facilities for extracting the data files from the bundled exe at runtime.
Instead, the setup.py process needs to ask for the package data files so
it can bundle them separately from the exe.  When the application is running
it will need to find the resource files so that it can load them, regardless
if it is running directly from the source tree, from an installed package
or from an exe or app.

The Resources class handles both the setup and the runtime requirements
for package resources.  You will need to put the resources in a path
within your source tree and initialize a Resources object with the necessary
information.  For example, assuming the resources are in a subdirectory
named "resources", and this package is stored as resfiles.py in the parent,
you would set resources/__init__.py as follows::

    from ..resfiles import Resources
    resources = Resources(package=__name__,
                          patterns=('*.png', '*.jpg', '*.ico', '*.wav'),
                          datadir='bumps-data',
                          check_file='reload.png',
                          env='BUMPS_DATA')

Now a resource file such as 'reload.png' can be accessed from a parent module
using::

    from .resources import resources
    resources.get_path('reload.png')

"""

import sys
import os
import glob


class Resources(object):
    """
    Identify project resource files.

    *package* : string
        Name of the subpackage containing the resources.  From the __init__.py
        file, for the resource directory, this is just __name__.
    *patterns* : list or tuple
        Set of glob patterns used to identify resource files in the resource
        package.
    *datadir* : string
        Name of the installed resource directory.  This is used in setup to
        prepare the resource directory and in the application to locate the
        resources that have been installed.
    *check_file*: string
        Name of a resource file that should be in the resource directory.  This
        is used to check that the resource directory exists in the installed
        application.
    *env* : string (optional)
        Environment variable which contains the complete path to the resource
        directory.  The environment variable overrides other methods of
        accessing the resources except running directly from the source tree.
    """

    def __init__(self, package, patterns, datadir, check_file, env=None):
        self.package = package
        self.patterns = patterns
        self.datadir = datadir
        self.check_file = check_file
        self.env = env
        self._cached_path = None

    def package_data(self):
        """
        Return the data files associated with the package.

        The format is a dictionary of {'fully.qualified.package', [files...]}
        used directly in the setup script as::

            setup(...,
                  package_data=package_data(),
                  ...)
        """
        return {self.package: list(self.patterns)}

    def data_files(self):
        """
        Return the data files associated with the package.

        The format is a list of (directory, [files...]) pairs which can be
        used directly in the py2exe setup script as::

            setup(...,
                  data_files=data_files(),
                  ...)

        Unlike package_data(), which only works from the source tree, data_files
        uses installed data path to locate the resources.
        """
        data_files = [(self.datadir, self._finddata(*self.patterns))]
        return data_files

    def _finddata(self, *patterns):
        path = self.resource_dir()
        files = []
        for p in patterns:
            files += glob.glob(os.path.join(path, p))
        return files

    def resource_dir(self):
        """
        Return the path to the application data.

        This is either in an environment variable, in the source tree next to
        this file, or beside the executable.  The environment key can be set
        using
        """
        # If we already found it, then we are done
        if self._cached_path is not None:
            return self._cached_path

        # Check for data in the package itself (which will be the case when
        # we are running from the source tree).
        path = os.path.abspath(os.path.dirname(sys.modules[self.package].__file__))
        if self._cache_resource_path(path):
            return self._cached_path

        # Check for data path in the environment.  If the environment variable
        # is specified, then the resources have to be there, or the program fails.
        if self.env and self.env in os.environ:
            if not self._cache_resource_path(os.environ[self.env]):
                raise RuntimeError("Environment %s not a directory" % self.env)
            return self._cached_path

        # Check for data next to exe/zip file.
        exepath = os.path.dirname(sys.executable)
        path = os.path.join(exepath, self.datadir)
        if self._cache_resource_path(path):
            return self._cached_path

        # py2app puts the data in Contents/Resources, but the executable
        # is in Contents/MacOS.
        path = os.path.join(exepath, "..", "Resources", self.datadir)
        if self._cache_resource_path(path):
            return self._cached_path

        raise RuntimeError("Could not find the GUI data files")

    def _cache_resource_path(self, path):
        if os.path.exists(os.path.join(path, self.check_file)):
            self._cached_path = path
            return True
        else:
            return False

    def get_path(self, filename):
        return os.path.join(self.resource_dir(), filename)



"""
Interface between the models and the fitters.

:class:`Fitness` defines the interface that model evaluators can follow.
These models can be bundled together into a :func:`FitProblem` and sent
to :class:`bumps.fitters.FitDriver` for optimization and uncertainty
analysis.


Summary of problem attributes::

    # Used by fitters
    nllf(p: Optional[Vector]) -> float  # main calculation
    bounds() -> Tuple(Vector, Vector)    # or equivalent sequence
    setp(p: Vector) -> None
    getp() -> Vector
    residuals() -> Vector  # for LM, MPFit
    parameter_residuals() -> Vector  # for LM, MPFit
    constraints_nllf() -> float # for LM, MPFit;  constraint cost is spread across the individual residuals
    randomize() -> None # for multistart
    resynth_data() -> None  # for Monte Carlo resampling of maximum likelihood
    restore_data() -> None # for Monte Carlo resampling of maximum likelihood
    name: str  # DREAM uses this
    chisq() -> float
    chisq_str() -> str
    labels() -> List[str]
    summarize() -> str
    show() -> None
    load(input_path: str) -> None
    save(output_path: str) -> None
    plot(figfile: str, view: str) -> None

    # Set/used by bumps.cli
    model_reset() -> None # called by load_model
    path: str  # set by load_model
    name: str # set by load_model
    title: str = filename # set by load_moel
    options: List[str]  # from sys.argv[1:]
    undefined:List[int]  # when loading a save .par file, these parameters weren't defined
    store: str # set by make_store
    output_path: str # set by make_store
    simulate_data(noise: float) -> None # for --simulate in opts
    cov() -> Matrix   # for --cov in opts

"""

__all__ = ["Fitness", "FitProblem", "load_problem"]

from dataclasses import dataclass
import logging
import os
import sys
import traceback
from typing import Generic, TypeVar
import warnings

import numpy as np
from numpy import inf, isnan, nan

from . import parameter, util
from .parameter import to_dict, Parameter, Variable, tag_all
from .formatnum import format_uncertainty


# Abstract base class:
# can use "isinstance" to check if a class implements the protocol
@util.runtime_checkable
@dataclass(init=False)
class Fitness(util.Protocol):
    """
    Manage parameters, data, and theory function evaluation.

    See :ref:`fitness` for a detailed explanation.
    """

    def parameters(self) -> util.List[Parameter]:
        """
        return the parameters in the model.

        model parameters are a hierarchical structure of lists and
        dictionaries.
        """
        raise NotImplementedError()

    def update(self):
        """
        Called when parameters have been updated.  Any cached values will need
        to be cleared and the model reevaluated.
        """
        raise NotImplementedError()

    def numpoints(self):
        """
        Return the number of data points.
        """
        raise NotImplementedError()

    def nllf(self):
        """
        Return the negative log likelihood value of the current parameter set.
        """
        raise NotImplementedError()

    def resynth_data(self):
        """
        Generate fake data based on uncertainties in the real data.  For
        Monte Carlo resynth-refit uncertainty analysis.  Bootstrapping?
        """
        raise NotImplementedError()

    def restore_data(self):
        """
        Restore the original data in the model (after resynth).
        """
        raise NotImplementedError()

    def residuals(self):
        """
        Return residuals for current theory minus data.

        Used for Levenburg-Marquardt, and for plotting.
        """
        raise NotImplementedError()

    def save(self, basename):
        """
        Save the model to a file based on basename+extension.  This will point
        to a path to a directory on a remote machine; don't make any
        assumptions about information stored on the server.  Return the set of
        files saved so that the monitor software can make a pretty web page.
        """
        pass

    def plot(self, view="linear"):
        """
        Plot the model to the current figure.  You only get one figure, but you
        can make it as complex as you want.  This will be saved as a png on
        the server, and composed onto a results web page.
        """
        pass


def no_constraints() -> float:
    """default constraints function for FitProblem"""
    return 0


def fit_parameters(fitness: Fitness) -> util.List[Parameter]:
    """
    Return a list of fittable (non-fixed) parameters in the model
    """
    parameters = parameter.unique(fitness.parameters())
    return [p for p in parameters if isinstance(getattr(p, "slot", None), parameter.Variable) and not p.fixed]


def chisq_str(fitness: Fitness) -> str:
    """
    Return a string representing the chisq equivalent of the nllf.
    If the model has strictly gaussian independent uncertainties then the
    negative log likelihood function will return 0.5*sum(residuals**2),
    which is 1/2*chisq.  Since we are printing normalized chisq, we
    multiply the model nllf by 2/DOF before displaying the value.  This
    is different from the problem nllf function, which includes the
    cost of the prior parameters and the cost of the penalty constraints
    in the total nllf.  The constraint value is displayed separately.
    """

    pars = fit_parameters(fitness)
    dof = fitness.numpoints() - len(pars)
    if dof <= 0 or np.isnan(dof) or np.isinf(dof):
        chisq_norm, chisq_err = 1.0, 0.0
    else:
        # return 2./dof, 1./dof
        from scipy.stats import chi2

        npars = max(len(pars), 1)
        chisq_norm, chisq_err = 2.0 / dof, chi2.ppf(ONE_SIGMA, npars) / dof

    chisq = fitness.nllf() * chisq_norm
    text = format_uncertainty(chisq, chisq_err)
    return text


def show_parameters(fitness: Fitness, subs: util.Optional[util.Dict[util.Any, Parameter]] = None):
    """Print the available parameters to the console as a tree."""
    print(parameter.format(fitness.parameters(), freevars=subs))
    print("[chisq=%s, nllf=%g]" % (chisq_str(fitness), fitness.nllf()))


FitnessType = TypeVar("FitnessType", bound=Fitness)


@dataclass(init=False, eq=False)
class FitProblem(Generic[FitnessType]):
    r"""

        *models* is a sequence of :class:`Fitness` instances.

        *weights* is an optional scale factor for each model. A weighted fit
        returns nllf $L = \sum w_k^2 L_k$. If an individual nllf is the sum
        squared residuals then this is equivalent to scaling the measurement
        uncertainty by $1/w$. Unless the measurement uncertainty is unknown,
        weights should be in [0, 1], representing an unknown systematic
        uncertainty spread across the individual measurements.

        *freevars* is :class:`.parameter.FreeVariables` instance defining the
        per-model parameter assignments.  See :ref:`freevariables` for details.


    Additional parameters:

        *name* name of the problem

        *constraints* is a list of Constraint objects, which have a method to
        calculate the nllf for that constraint.
        Also supports an alternate form which cannot be serialized:
        A function which returns the negative log likelihood
        of seeing the parameters independent from the fitness function.  Use
        this for example to check for feasible regions of the search space, or
        to add constraints that cannot be easily calculated per parameter.
        Ideally, the constraints nllf will increase as you go farther from
        the feasible region so that the fit will be directed toward feasible
        values.

        *penalty_nllf* is the nllf to use for *fitness* when *constraints*
        or model parameter bounds are not satisfied.

    Total nllf is the sum of the parameter nllf, the constraints nllf and the
    depending on whether constraints is greater than soft_limit, either the
    fitness nllf or the penalty nllf.

    New in 0.9.0: weights are now squared when computing the sum rather than
    linear.
    """

    name: util.Optional[str]
    models: util.List[FitnessType]
    freevars: util.Optional[parameter.FreeVariables]
    weights: util.Union[util.List[float], util.Literal[None]]
    constraints: util.Optional[util.Sequence[parameter.Constraint]]
    penalty_nllf: util.Union[float, util.Literal["inf"]]

    _constraints_function: util.Callable[..., float]
    _models: util.List[FitnessType]
    _parameters: util.List[Parameter]
    _parameters_by_id: util.Dict[str, Parameter]
    _dof: float = np.nan  # not a schema field, and is not used in __init__

    # _all_constraints: util.List[util.Union[Parameter, Expression]]

    def __init__(
        self,
        models: util.Union[FitnessType, util.List[FitnessType]],
        weights=None,
        name=None,
        constraints=None,
        penalty_nllf="inf",
        freevars=None,
        soft_limit: util.Optional[float] = None,  # TODO: deprecate,
        auto_tag=False,
    ):
        if not isinstance(models, (list, tuple)):
            models = [models]
        if callable(constraints):
            warnings.warn("convert constraints functions to constraint expressions", DeprecationWarning)
            self._constraints_function = constraints
            self.constraints = []
        else:
            self._constraints_function = self._null_constraints_function
            # TODO: do we want to allow "constraints=a<b" or do we require a sequence "constraints=[a<b]"?
            self.constraints = constraints if constraints is not None else []
        if freevars is None:
            names = ["M%d" % i for i, _ in enumerate(models)]
            freevars = parameter.FreeVariables(names=names)
        self.freevars = freevars
        if auto_tag:
            for index, model in enumerate(models):
                model_name = model.name if model.name is not None else f"Model{index}"
                tag_all(model.parameters(), model_name)
        self._models = models
        if weights is None:
            weights = [1.0 for _ in models]
        self.weights = weights
        self.penalty_nllf = float(penalty_nllf)
        self.set_active_model(0)  # Set the active model to model 0
        self.model_reset()  # sets self._all_constraints
        self.name = name

    @staticmethod
    def _null_constraints_function():
        return 0.0

    @property
    def fitness(self):
        warnings.warn("Deprecated: use of problem.fitness will be removed at some point")
        if len(self._models) == 1:
            return self._models[0]
        raise ValueError("problem.fitness is not defined")

    @property
    def dof(self):
        return self._dof

    # TODO: make this @property\ndef models(self): ...
    @property
    def models(self):
        """Iterate over models, with free parameters set from model values"""
        try:
            for i, f in enumerate(self._models):
                self.freevars.set_model(i)
                yield f
        finally:
            # Restore the active model after cycling, even if interrupted
            self.freevars.set_model(self._active_model_index)

    # noinspection PyAttributeOutsideInit
    def set_active_model(self, i):
        """Use free parameters from model *i*"""
        self._active_model_index = i
        self.active_model = self._models[i]
        self.freevars.set_model(i)

    def model_parameters(self):
        """Return parameters from all models"""
        pars = {}
        pars["models"] = [f.parameters() for f in self.models]
        free = self.freevars.parameters()
        if free:
            pars["freevars"] = free
        return pars

    def to_dict(self):
        return {
            "type": type(self).__name__,
            "name": self.name,
            "models": to_dict(self._models),
            "weights": self.weights,
            "penalty_nllf": self.penalty_nllf,
            # TODO: constraints may be a function.
            "constraints": to_dict(self.constraints),
            "freevars": to_dict(self.freevars),
        }

    def __repr__(self):
        return "FitProblem(name=%s)" % self.name

    def valid(self, pvec):
        """Return true if the point is in the feasible region"""
        return all(v in p.prior for p, v in zip(self._parameters, pvec))

    def setp(self, pvec):
        """
        Set a new value for the parameters into the model.  If the model
        is valid, calls model_update to signal that the model should be
        recalculated.

        Returns True if the value is valid and the parameters were set,
        otherwise returns False.
        """
        # print("Calling setp with", pvec, self._parameters)
        # TODO: do we have to leave the model in an invalid state?
        # WARNING: don't try to conditionally update the model
        # depending on whether any model parameters have changed.
        # For one thing, the model_update below probably calls
        # the subclass FitProblem.model_update, which signals
        # the individual models.  Furthermore, some parameters may
        # related to others via expressions, and so a dependency
        # tree needs to be generated.  Whether this is better than
        # clicker() from SrFit I do not know.
        for v, p in zip(pvec, self._parameters):
            p.value = v
        # TODO: setp_hook is a hack to support parameter expressions in sasview
        # Don't depend on this existing long term.
        setp_hook = getattr(self, "setp_hook", no_constraints)
        setp_hook()
        self.model_update()

    def getp(self):
        """
        Returns the current value of the parameter vector.
        """
        return np.array([p.value for p in self._parameters], "d")

    def bounds(self):
        """Return the bounds for each parameter as a 2 x N array"""
        limits = [p.prior.limits for p in self._parameters]
        return np.array(limits, "d").T if limits else np.empty((2, 0))

    def randomize(self, n=None):
        """
        Generates a random model.

        *randomize()* sets the model to a random value.

        *randomize(n)* returns a population of *n* random models.

        For indefinite bounds, the random population distribution is centered
        on initial value of the parameter, or 1. if the initial parameter is
        not finite.
        """
        # TODO: split into two: randomize and random_pop
        if n is None:
            self.setp(self.randomize(n=1)[0])
            return  # Not returning anything since no n is requested

        # TODO: apply hard limits on parameters
        target = self.getp()
        target[~np.isfinite(target)] = 1.0
        pop = [p.prior.random(n, target=v) for p, v in zip(self._parameters, target)]
        return np.array(pop).T

    def nllf(self, pvec=None) -> float:
        """
        compute the cost function for a new parameter set p.

        this is not simply the sum-squared residuals, but instead is the
        negative log likelihood of seeing the data given the model parameters
        plus the negative log likelihood of seeing the model parameters.  the
        value is used for a likelihood ratio test so normalization constants
        can be ignored.  there is an additional penalty value provided by
        the model which can be used to implement inequality constraints.  any
        penalty should be large enough that it is effectively excluded from
        the parameter space returned from uncertainty analysis.

        the model is not actually calculated if any of the parameters are
        out of bounds, or any of the constraints are not satisfied, but
        instead are assigned a value of *penalty_nllf*.
        this will prevent expensive models from spending time computing
        values in the unfeasible region.
        """
        if pvec is not None:
            if self.valid(pvec):
                self.setp(pvec)
            else:
                return inf

        pparameter, pconstraints, pmodel, failing_constraints = self._nllf_components()
        cost = pparameter + pconstraints + pmodel
        # print("cost=",pparameter,"+",pconstraints,"+",pmodel,"=",cost, pvec)
        if isnan(cost):
            # TODO: make sure errors get back to the user
            # print "point evaluates to nan"
            # print parameter.summarize(self._parameters)
            return inf
        return cost

    def parameter_nllf(self) -> util.Tuple[float, util.List[str]]:
        """
        Returns negative log likelihood of seeing parameters p.
        """
        failing = []
        nllf = 0.0
        for p in self._bounded:
            p_nllf = p.nllf()
            nllf += p_nllf
            if p_nllf == np.inf:
                failing.append(str(p))
        # print "; ".join("%s %g %g"%(p,p.value,p.nllf()) for p in
        # self._bounded)
        return nllf, failing

    def parameter_residuals(self):
        """
        Returns negative log likelihood of seeing parameters p.
        """
        return [p.residual() for p in self._bounded]

    def chisq_str(self):
        """
        Return a string representing the chisq equivalent of the nllf.

        If the model has strictly gaussian independent uncertainties then the
        negative log likelihood function will return 0.5*sum(residuals**2),
        which is 1/2*chisq.  Since we are printing normalized chisq, we
        multiply the model nllf by 2/DOF before displaying the value.  This
        is different from the problem nllf function, which includes the
        cost of the prior parameters and the cost of the penalty constraints
        in the total nllf.  The constraint value is displayed separately.
        """
        pparameter, pconstraints, pmodel, failing_constraints = self._nllf_components()
        chisq_norm, chisq_err = nllf_scale(self)
        chisq = pmodel * chisq_norm
        text = format_uncertainty(chisq, chisq_err)
        constraints = pparameter + pconstraints
        if constraints > 0.0:
            text += " constraints=%g" % constraints
        if len(failing_constraints) > 0:
            text += " failing_constraints=%s" % str(failing_constraints)

        return text

    def _nllf_components(self) -> util.Tuple[float, float, float, util.List[str]]:
        try:
            pparameter, failing_parameter_constraints = self.parameter_nllf()
            if isnan(pparameter):
                # TODO: make sure errors get back to the user
                info = ["Parameter nllf is wrong"]
                info += ["%s %g" % (p, p.nllf()) for p in self._bounded]
                logging.error("\n  ".join(info))
            pconstraints, failing_constraints = self.constraints_nllf()
            failing_constraints += failing_parameter_constraints
            # Note: for hard constraints (which return inf) avoid computing
            # model even if soft_limit is inf by using strict comparison
            # since inf <= inf is True but inf < inf is False.
            penalty_nllf = self.penalty_nllf if self.penalty_nllf is not None else np.inf
            pmodel = self.model_nllf() if len(failing_constraints) == 0 else penalty_nllf

            return pparameter, pconstraints, pmodel, failing_constraints
        except Exception:
            # TODO: make sure errors get back to the user
            info = (traceback.format_exc(), parameter.summarize(self._parameters))
            logging.error("\n".join(info))
            return nan, nan, nan, []

    def __call__(self, pvec=None):
        """
        Problem cost function.

        Returns the negative log likelihood scaled by DOF so that
        the result looks like the familiar normalized chi-squared.  These
        scale factors will not affect the value of the minimum, though some
        care will be required when interpreting the uncertainty.
        """
        return 2 * self.nllf(pvec) / self.dof

    def summarize(self):
        """Return a table of current parameter values with range bars."""
        return parameter.summarize(self._parameters)

    def labels(self) -> util.List[str]:
        """Return the list of labels, one per fitted parameter."""
        return [p.name for p in self._parameters]

    # noinspection PyAttributeOutsideInit
    def model_reset(self):
        """
        Prepare for the fit.

        This sets the parameters and the bounds properties that the
        solver is expecting from the fittable object.  We also compute
        the degrees of freedom so that we can return a normalized fit
        likelihood.

        If the set of fit parameters changes, then model_reset must
        be called.
        """
        # print("In model reset with", self.model_parameters())
        all_parameters = parameter.unique(self.model_parameters())
        # print "all_parameters",all_parameters
        # for p in all_parameters:
        #     if hasattr(p, 'reset_prior'):
        #         p.reset_prior()  # no constraints
        #     else:
        #         raise ValueError(f"{p} does not have prior")
        broken = []
        for p in all_parameters:
            # slot = p.slot
            # value = p.value

            # TODO: this is a shim to accomodate Expression, Calculation etc. being
            # put into attributes that have type "Parameter" (in user scripts)
            # Do we cause those scripts to break instead?
            # Or do we autoconvert to .equals()?
            if hasattr(p, "add_prior"):
                p.add_prior(p.distribution, bounds=p.bounds, limits=p.limits)

            # TODO: currently this logic for showing breaking constraints to the user
            # is in the chisq_str execution path instead of on model_reset - should it be here instead?

            # While we are walking all parameters check which constraints aren't satisfied
            # Build up a list of strings to help the user initialize the model correctly
            if hasattr(p, "limits"):
                value = p.value
                if (p.limits[0] > value) or (value > p.limits[1]):
                    broken.append(f"{p}={value} is outside {p.limits}")
                elif not np.isfinite(p.prior.nllf(value)):
                    broken.append(f"{p}={value} is outside {p.prior}")

        broken.extend([f"constraint {c} is unsatisfied" for c in self.constraints if float(c) == inf])
        if self._constraints_function() == inf:
            broken.append("user constraint function is unsatisfied")
        if len(broken) > 0:
            warnings.warn("Unsatisfied constraints: [%s]" % (",\n".join(broken)))
        self.broken_constraints = broken

        # TODO: shimmed to allow non-Parameter in Parameter attribute spots.
        pars = []
        pars_by_id = {}
        for p in all_parameters:
            if hasattr(p, "id"):
                pars_by_id[p.id] = p
            slot = getattr(p, "slot", None)
            if isinstance(slot, Variable) and not p.fixed:
                pars.append(p)
        self._parameters = pars
        self._parameters_by_id = pars_by_id
        # TODO: shimmed to allow non-Parameter in Parameter attribute spots.
        self._bounded = [p for p in all_parameters if hasattr(p, "has_prior") and p.has_prior()]
        self._dof = self.model_points()
        self._dof -= len(self._parameters)
        if self.dof <= 0:
            warnings.warn(
                f"Need more data points (currently: {self.model_points()}) than fitting parameters ({len(self._parameters)})"
            )
        # self.constraints = pars.constraints()
        # Find the constraints on variables and expressions that we need to compute
        # parameter_constraints = [p.slot for p in all_parameters if isinstance(p.slot, Variable) and p.has_prior()]
        # expression_constraints = [p.slot for p in all_parameters if isinstance(p.slot, Expression) and p.has_prior()]
        # self._all_constraints = parameter_constraints + expression_constraints

    def model_points(self):
        """Return number of points in all models"""
        return sum(f.numpoints() for f in self.models)

    def model_update(self):
        """Let all models know they need to be recalculated"""
        # TODO: consider an "on changed" signal for model updates.
        # The update function would be associated with model parameters
        # rather than always recalculating everything.  This
        # allows us to set up fits with 'fast' and 'slow' parameters,
        # where the fit can quickly explore a subspace where the
        # computation is cheap before jumping to a more expensive
        # subspace.  SrFit does this.
        for f in self.models:
            if hasattr(f, "update"):
                f.update()

    def model_nllf(self):
        """Return cost function for all data sets"""
        return sum(w**2 * f.nllf() for w, f in zip(self.weights, self.models))

    def constraints_nllf(self) -> util.Tuple[float, util.List[str]]:
        """Return the cost function for all constraints"""
        failing = []
        nllf = 0.0
        nllf = self._constraints_function()
        if nllf == np.inf:
            failing.append("user constraints function")
        for c in self.constraints:
            # TODO: convert to list of residuals for Levenberg-Marquardt
            c_nllf = float(c) ** 2
            nllf += c_nllf
            if c_nllf > 0:
                failing.append(str(c))

        return nllf, failing

    def simulate_data(self, noise=None):
        """Simulate data with added noise"""
        for f in self.models:
            f.simulate_data(noise=noise)

    def resynth_data(self):
        """Resynthesize data with noise from the uncertainty estimates."""
        for f in self.models:
            f.resynth_data()

    def restore_data(self):
        """Restore original data after resynthesis."""
        for f in self.models:
            f.restore_data()

    @property
    def has_residuals(self):
        """
        True if all underlying fitness functions define residuals.
        """
        return all(hasattr(f, "residuals") for f in self.models)

    def residuals(self):
        resid = np.hstack([w * f.residuals() for w, f in zip(self.weights, self.models)])
        return resid

    def save(self, basename):
        for i, f in enumerate(self.models):
            if hasattr(f, "save"):
                f.save(basename + "-%d" % (i + 1))

    def show(self):
        for i, f in enumerate(self.models):
            print("-- Model %d %s" % (i, getattr(f, "name", "")))
            subs = self.freevars.get_model(i) if self.freevars else {}
            show_parameters(f, subs=subs)
        print("[overall chisq=%s, nllf=%g]" % (self.chisq_str(), self.nllf()))

    def plot(self, p=None, fignum=1, figfile=None, view=None, model_indices=None):
        import pylab

        if p is not None:
            self.setp(p)
        for i, f in enumerate(self.models):
            if model_indices is not None and i not in model_indices:
                continue
            if not hasattr(f, "plot"):
                continue
            f.plot(view=view)
            pylab.figure(i + fignum)
            f.plot(view=view)
            pylab.suptitle("Model %d - %s" % (i, f.name))
            pylab.text(0.01, 0.01, "chisq=%s" % chisq_str(f), transform=pylab.gca().transAxes)
            if figfile is not None:
                pylab.savefig(figfile + "-model%d.png" % i, format="png")

    # Note: restore default behaviour of getstate/setstate rather than
    # inheriting from BaseFitProblem
    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state


# TODO: consider adding nllf_scale to FitProblem.
ONE_SIGMA = 0.68268949213708585


def nllf_scale(problem: FitProblem):
    r"""
    Return the scale factor for reporting the problem nllf as an approximate
    normalized chisq, along with an associated "uncertainty".  The uncertainty
    is the amount that chisq must change in order for the fit to be
    significantly better.

    From Numerical Recipes 15.6: *Confidence Limits on Estimated Model
    Parameters*, the $1-\sigma$ contour in parameter space corresponds
    to $\Delta\chi^2 = \text{invCDF}(1-\sigma,k)$ where
    $1-\sigma \approx 0.6827$ and $k$ is the number of fitting parameters.
    Since we are reporting the normalized $\chi^2$, this needs to be scaled
    by the problem degrees of freedom, $n-k$, where $n$ is the number of
    measurements.  To first approximation, the uncertainty in $\chi^2_N$
    is $k/(n-k)$
    """
    dof = problem.dof
    if dof <= 0 or np.isnan(dof) or np.isinf(dof):
        return 1.0, 0.0
    else:
        # return 2./dof, 1./dof
        from scipy.stats import chi2

        npars = max(len(problem.getp()), 1)
        return 2.0 / dof, chi2.ppf(ONE_SIGMA, npars) / dof


def load_problem(filename, options=None) -> FitProblem:
    """
    Load a problem definition from a python script file.

    sys.argv is set to ``[file] + options`` within the context of the script.

    The user must define ``problem=FitProblem(...)`` within the script.

    Raises ValueError if the script does not define problem.
    """
    # Allow relative imports from the bumps model
    module_name = os.path.splitext(os.path.basename(filename))[0]
    module = util.relative_import(filename, module_name=module_name)

    ctx = dict(__file__=filename, __package__=module, __name__=module_name)
    old_argv = sys.argv
    sys.argv = [filename] + options if options else [filename]
    source = open(filename).read()
    code = compile(source, filename, "exec")
    exec(code, ctx)
    sys.argv = old_argv
    problem = ctx.get("problem", None)
    if problem is None:
        raise ValueError(filename + " requires 'problem = FitProblem(...)'")

    return problem


def MultiFitProblem(*args, **kwargs) -> FitProblem:
    warnings.warn(DeprecationWarning("use FitProblem directly instead of MultiFitProblem"))
    return FitProblem(*args, **kwargs)


def test_weighting():
    class SimpleFitness(Fitness):
        def __init__(self, a=0.0, name="fit"):
            self.a = parameter.Parameter.default(a, name=name + " a")

        def parameters(self):
            return {"a": self.a}

        def numpoints(self):
            return 1

        def residuals(self):
            y, dy = 0, 1  # fit to constant 0 +/- 1
            return np.array([(self.a.value - y) / dy])

        def nllf(self):
            return sum(r**2 for r in self.residuals()) / 2

    weights = 2, 3
    models = [SimpleFitness(4.0), SimpleFitness(5.0)]
    problem = FitProblem(models, weights=weights)

    # Need to use problem.models to cycle through models in case FreeVariables is used in problem
    assert (problem.residuals() == np.hstack([w * M.residuals() for w, M in zip(weights, problem.models)])).all()
    assert problem.nllf() == sum(w**2 * M.nllf() for w, M in zip(weights, problem.models))
    assert problem.nllf() == sum(problem.residuals() ** 2) / 2



r"""
Convergence diagnostics
=======================

The function :func:`burn_point` returns the point within the MCMC
chain at which the chain can be said to have converged, or -1 if
the log probabilities are still improving throughout the chain.
"""

__all__ = ["burn_point", "ks_converged"]

from typing import TYPE_CHECKING

import numpy as np
from numpy.random import choice
from scipy.stats import chi2, ks_2samp, kstest

if TYPE_CHECKING:
    from .state import MCMCDraw

# TODO is cramer von mises better than a KS test?

# defaults should match for ks_converged and burn_point trimming
DENSITY = 0.6
ALPHA = 0.01
SAMPLES = 1000
TRIALS = 5
MIN_WINDOW = 100


def ks_converged(state, trials=TRIALS, density=DENSITY, alpha=ALPHA, samples=SAMPLES):
    # type: ("MCMCDraw", int, float, float, int) -> bool
    """
    Return True if the MCMC has converged according to the K-S window test.

    Since we are only looking at the distribution of logp values, and not the
    individual points, we should be relatively stable regardless of the
    properties of the sampler.  The main reason for failure will be "stuck"
    fits which have not had a chance to jump to a lower minimum.

    *state* contains the MCMC chain information.

    *trials* is the number of times to run the K-S test.

    *density* is the proportion of samples to select from the window.  Prefer
    lower density from larger number of samples so the sets chosen for the
    K-S test have fewer duplicates.  For density=0.1 about 5% of samples will
    be duplicates.  For density=0.6 about 25% will be duplicates.

    *alpha* is the significance level for the test.  With smaller alpha
    values the K-S test is less likely to reject the current window when
    testing against the tail of the distribution, and so the fit will end
    earlier, with more samples after the burn point.

    *samples* is the size of the sample window. If the window is too big
    the test will falsly end burn when the start of the window is still
    converging.  If the window is too small the test will take a long time,
    and will start to show effects of autocorrelation (efficient MCMC
    samplers move slowly across the posterior probability space, showing
    short term autocorrelation between samples.)  A minimum of 10 generations
    and a maximum of 1/2 the generations will be used.

    There is a strong interaction between density, alpha, samples and trials.
    If the K-S test has too many points (=density*samples), it will often
    reject simply because the different portions of the Markov chain are
    not identical (Markov chains can have short range correlations yet
    still have the full chain as a representative draw from the posterior
    distribution) unless alpha is reduced.  With fewer points, the estimated
    K-S statistic will have more variance, and so more trials will be needed
    to avoid spurious accept/reject decisions.
    """
    # Make sure we are testing for convergence
    if alpha == 0.0:
        return False

    # Make sure we have the desired number of draws
    if state.generation < state.Ngen:
        return False

    # Quick fail if best occurred within draw
    if not state.stable_best():
        # print(state.generation, "best gen", state._best_gen, "start", state.generation - state.Ngen)
        return False

    # Grab a window at the start and the end
    window_size = min(max(samples // state.Npop + 1, MIN_WINDOW), state.Ngen // 2)

    head = state.logp_slice(window_size).flatten()
    tail = state.logp_slice(-window_size).flatten()

    # Quick fail if logp head is worse than logp tail
    if np.min(head) < state.min_slice(-state.Ngen // 2):
        # print(state.generation, "improving worst", np.min(head), state.min_slice(-state.Ngen//2))
        return False

    n_draw = int(density * samples)
    reject = _robust_ks_2samp(head, tail, n_draw, trials, alpha)
    if reject:
        return False

    return True


def check_nllf_distribution(state):
    """
    Check if the nllf distribution looks like chisq.

    Note: test is not used.  It is only true for gaussian.  It fails pretty
    badly for doc/examples/test_functions.py griewank 2.  It fails even
    worse with pure integer models such as the OpenBugs asia example, which
    has discrete levels in the posterior pdf corresponding to the various
    binary configurations of the model values.
    """

    # Check that likelihood distribution looks like chi2
    # Note: cheating, and looking at the stored logp without unrolling
    reject = _check_nllf_distribution(data=-state._gen_logp.flatten(), df=state.Nvar, n_draw=10, trials=5, alpha=0.01)
    return not reject


def _check_nllf_distribution(data, df, n_draw, trials, alpha):
    # fit the best chisq to the data given df
    float_df, loc, scale = chi2.fit(data, f0=df)
    df = int(float_df + 0.5)
    cdf = lambda x: chi2.cdf(x, df, loc, scale)

    # check the quality of the fit (i.e., does the set of nllfs look vaguely
    # like the fitted chisq distribution).  Repeat the test a few times on
    # small data sets for consistency.
    p_vals = []
    for _ in range(trials):
        f_samp = choice(data, n_draw, replace=True)
        p_vals.append(kstest(data, cdf)[1])

    print("llf dist", p_vals, df, loc, scale)
    return alpha > np.mean(p_vals)


def burn_point(state, method="window", trials=TRIALS, **kwargs):
    # type: ("MCMCDraw", str, int, **dict) -> int

    r"""
    Determines the point at which the MCMC chain seems to have converged.

    *state* contains the MCMC chain information.

    *method="window"* is the name of the convergence diagnostic (see below).

    *trials* is the number of times to run the K-S test.

    Returns the index of the burn points, or -1 if no good burn point is found.

    **Kolmogorov-Smirnov sliding window**

    The "window" method detects convergence by comparing the distribution of
    $\log(p)$ values in a small window at the start of the chains and the
    values in a section at the end of the chain using a Kolmogorov-Smirnov
    test.  See :func:`ks_converged` for a description of the parameters.
    """
    if method == "window":
        index = _ks_sliding_window(state, trials=trials, **kwargs)
    else:
        raise ValueError("Unknown convergence test " + method)
    if index < 0:
        print("Did not converge!")
    return index


def _ks_sliding_window(state, trials=TRIALS, density=DENSITY, alpha=ALPHA * 0.01, samples=SAMPLES):
    _, logp = state.logp()

    window_size = min(max(samples // state.Npop + 1, MIN_WINDOW), state.Ngen // 2)
    tiny_window = window_size // 11 + 1
    half = state.Ngen // 2
    max_index = len(logp) - half - window_size

    if max_index < 0:
        return -1
    tail = logp[-half:].flatten()
    min_tail = np.min(tail)

    # Check in large bunches
    n_draw = int(density * samples)
    for index in range(0, max_index + 1, window_size):
        # [PAK] make sure the worst point is not in the first window.
        # Stastically this will introduce some bias (by chance the max could
        # happen to occur in the first window) but it will be small when the
        # window is small relative to the full pool.  A better test would
        # count the number of samples worse than the all the tail, compute
        # the probability, and reject according to a comparison with a uniform
        # number in [0,1].  To much work for so little bias.
        window = logp[index : index + window_size].flatten()
        if np.min(window) < min_tail:
            # print("step llf", index, window_size, len(window), np.min(window), min_tail)
            continue

        # if head and tail are different, slide to the next window
        reject = _robust_ks_2samp(window, tail, n_draw, trials, alpha)
        if reject:
            continue

        # Head and tail are not significantly different, so break.
        # Index is not yet updated, so the tiny step loop will start with
        # the first rejected window.
        break

    if index >= max_index:
        return -1

    # check in smaller steps for fine tuned stopping
    for index in range(index, index + window_size, tiny_window):
        window = logp[index : index + tiny_window].flatten()
        if np.min(window) < min_tail:
            # print("tiny llf", index, tiny_window, len(window), np.min(window), min_tail)
            continue

        p_val = _robust_ks_2samp(window, tail, n_draw, trials, alpha)
        # print("tiny ks", index, tiny_window, len(window), p_val, alpha)
        if p_val > alpha:
            # head and tail are not significantly different, so break
            break

    return index


def _robust_ks_2samp(f_data, r_data, n_draw, trials, alpha):
    """
    Repeat ks test n times for a more robust statistic.

    Returns True if f_data is significantly different from r_data.
    """
    p_vals = []
    for _ in range(trials):
        # [PAK] Using replace=True since it is more efficient and practically
        # indistinguishable for a small sampling portion such as 10% or less.
        # When drawing 60% of the sample size, 25% of the samples are repeated.
        f_samp = choice(f_data, n_draw, replace=True)
        r_samp = choice(r_data, n_draw, replace=True)
        p_vals.append(ks_2samp(f_samp, r_samp)[1])
    return alpha > np.mean(p_vals)
    # return any(alpha > p for p in p_vals)



"""
Miscellaneous utilities.
"""

__all__ = ["draw", "console"]

import numpy as np
import numpy.random as rng

try:
    from numba import njit
except ImportError:

    def njit(*args, **kw):
        return lambda f: f


@njit(cache=True)
def draw(k, n):
    """
    Select k things from a pool of n without replacement.
    """
    # At k == n/4, an extra 0.15*k draws are needed to get k unique draws
    # TODO: silently returns too few values if k > n
    if k > n / 4:
        result = rng.permutation(n)[:k]
    else:
        result = np.empty(k, np.int64)
        for i in range(k):
            # select an item not already selected
            while True:
                p = rng.randint(n)
                for j in range(i):
                    # if there is a match then break out of the for-loop and
                    # go to the next while iteration, generating a new proposal
                    if j == p:
                        break
                # unusual syntax: if you make it to the end of the for loop
                # then there were no matches, so break out of the while loop
                else:
                    break
            result[i] = p
    return result


def _check_uniform_draw():
    """
    Draws from history should
    """
    import pylab

    k, n = 50, 400
    counts = np.zeros(n * k)
    idx = np.arange(k)
    for _ in range(100000):
        t = draw(k, n)
        counts[k * t + idx] += 1
    pylab.subplot(211)
    pylab.pcolormesh(np.reshape(counts, (n, k)))
    pylab.colorbar()
    pylab.title("drawn number vs. draw position")
    pylab.subplot(212)
    pylab.hist(counts)
    pylab.title("number of draws per (number,position) bin")
    pylab.show()


def console():
    """
    Start the python console with the local variables available.

    console() should be the last thing in the file, after sampling and
    showing the default plots.
    """
    import os
    import sys

    # Hack for eclipse console: can't actually run ipython in the eclipse
    # console and get it to plot, so instead guess whether we are in a
    # console by checking if we are attached to a proper tty through stdin.
    # For eclipse, just show the plots.
    try:
        tty = os.isatty(sys.stdin.fileno())
    except Exception:
        tty = False

    if tty:
        # Display outstanding plots and turn interactive on
        from matplotlib import interactive
        from matplotlib._pylab_helpers import Gcf

        for fig in Gcf.get_all_fig_managers():
            try:  # CRUFT
                fig.show()
            except AttributeError:
                fig.frame.Show()
        interactive(True)

        # Start an ipython shell with the caller's local variables
        import IPython

        symbols = sys._getframe(1).f_locals
        ip = IPython.Shell.IPShell(user_ns=symbols)
        ip.mainloop()
    else:
        # Not a tty; try doing show() anyway
        import pylab

        pylab.show()



from collections import deque
from queue import Queue
from typing import Dict, List, Optional, TYPE_CHECKING
from threading import Event
import asyncio

from .state_hdf5_backed import SERIALIZERS

if TYPE_CHECKING:
    import bumps, bumps.fitproblem, bumps.dream.state
    from .webserver import TopicNameType
    from .fit_thread import FitThread


SESSION_FILE_NAME = "session.h5"


class ProblemState:
    fitProblem: Optional["bumps.fitproblem.FitProblem"] = None
    pathlist: Optional[List[str]] = None
    serializer: Optional[SERIALIZERS] = None
    filename: Optional[str] = None


class FittingState:
    abort: bool = False
    population: Optional[List] = None
    uncertainty_state: Optional["bumps.dream.state.MCMCDraw"]


class State:
    # These attributes are ephemeral, not to be serialized/stored:
    hostname: str
    port: int
    parallel: int
    abort_queue: Queue
    fit_thread: Optional["FitThread"] = None
    fit_abort: Optional[Event] = None
    fit_abort_event: Event
    fit_complete_event: Event
    calling_loop: Optional[asyncio.AbstractEventLoop] = None
    fit_enabled: Event
    session_file_name: Optional[str]

    # State to be stored:
    problem: ProblemState
    fitting: FittingState
    topics: Dict["TopicNameType", "deque[Dict]"]

    def __init__(self, problem: Optional[ProblemState] = None, fitting: Optional[FittingState] = None):
        self.problem = problem if problem is not None else ProblemState()
        self.fitting = FittingState()
        self.fit_abort_event = Event()
        self.fit_complete_event = Event()
        self.topics = {
            "log": deque([]),
            "update_parameters": deque([], maxlen=1),
            "update_model": deque([], maxlen=1),
            "model_loaded": deque([], maxlen=1),
            "fit_active": deque([], maxlen=1),
            "convergence_update": deque([], maxlen=1),
            "uncertainty_update": deque([], maxlen=1),
            "fitter_settings": deque([], maxlen=1),
            "fitter_active": deque([], maxlen=1),
        }

    def setup_backing(self, session_file_name: Optional[str] = SESSION_FILE_NAME, read_only: bool = False):
        self.session_file_name = session_file_name

    def save(self):
        pass

    def copy_session_file(self, session_copy_name: str):
        pass

    def write_session_file(self):
        pass

    def cleanup(self):
        self._session_file.close()

    def __del__(self):
        self.cleanup()

    async def async_cleanup(self):
        self.cleanup()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.cleanup()

    async def cleanup(self):
        pass



"""
Significant digits
------------------

Calculate the number significant digits in a statistic using boot
strapping/jackknife.

The 'gum' method from the GUM supplement[1] splits the sequence into n
independent sequences, computes the statistic for each sequence, and
returns the mean and variance of the statistic as the estimated value
with uncertainty.  Each block should have at least 10000 samples, or
100/(1-ci) samples if the desired credible interval ci > 0.99.

The 'jack' method uses

notes:

- the gum method gives a slight underestimate of the sd, while
  jackknife gives an overestimate.
- Jackknife performs better than gum on small samples
- in current implementation, gum is much faster
- in picking the number of cuts for jackknife, lower is usually better
  (but not too low...)

[1] JCGM, 2008. Evaluation of measurement data - Supplement 1 to the "Guide to
the expression of uncertainty in measurement" - Propagation of distributions
using a Monte Carlo method (No. JCGM 101:2008). Geneva, Switzerland.
"""

"""
THINGS TO DO:
- actually useful display of information in output
- an effective, fast, online version to use as a stoping crit
- use as a method to guess how many more samples:
    - provides estimate of sd of sampling distribution of statistic
    - sd of stat for n samples = sd for k samples / sqrt(n/k - 1)
    - not actually well tested...
- provide better user accessible control hooks
- do better deciding between jack and gum
- these don't deal with lists of CIs (as the actual ci functions do)
- very slow, needs a restructure
"""
# TODO: needs a refactor before going into production

import numpy as np
import numpy.ma as ma

from .stats import credible_interval


def credible_interval_sd(data, ci, fn=None, method=None, cuts=None):
    if fn is None:
        # ugly hack here
        fn = lambda *a, **kw: credible_interval(*a, **kw)[0]

    if method is None:
        # do something clever to choose gum or jack
        method = "gum"

    # Determine the number of cuts
    if cuts is None or cuts <= 0:
        # Use groups of 10000 unless looking for ci > 0.99, in which case
        # use groups of 100/(1-ci).
        M = max(int(100 / (1 - ci)), 10000)
        cuts = len(data) // M
    if cuts < 3:
        cuts = 3

    if method == "gum":
        return gum_sd(data, fn, ci)
    elif method == "jack":
        return jack_sd(data, fn, ci)
    else:
        raise ValueError("Unknown sd method" + method)


# TODO does not handle either CI function natively
# requires a 1d ndarray return value


def gum_sd(data, f, ci, cuts=10):
    """
    adaptation of method in section 7.9.4 of gum suppliment 1
    """
    # reshape data into blocks, skipping the first partial block
    M = len(data) // cuts
    data = data[-M * cuts :].reshape((cuts, M))
    # compute the variance of the statistic over the sub-blocks
    stat = np.apply_along_axis(f, 1, data, ci)
    var = np.var(stat, axis=0, ddof=1)
    # estimate standard deviation of the statistic for the full data set
    return np.sqrt(var / cuts)


def jack_sd(data, f, ci, cuts=10):
    """
    does an average of sd on smaller blocks to counteract skewed
    distribution of jackknife
    """
    # reshape data into blocks, skipping the first partial block
    M = len(data) // cuts
    data = data[-M * cuts :].reshape((cuts, M))
    # compute the standard deviation of the statistic over the sub-blocks
    std = np.apply_along_axis(fast_jack, 1, data, f, ci)
    # estimate standard deviation of the statistic for the full data set
    return np.mean(std, axis=0) / np.sqrt(cuts)


# TODO: does not deal well with a list of CIs
# returns the sd of a statistic, f
def fast_jack(data, f, *a, **kw):
    jack = FastJackknife(data=data, fn=f)
    jack.set_args(*a, **kw)
    return jack.std()


# TODO: this is coded badly and so is very slow
class FastJackknife(object):
    def __init__(self, **kw):
        self.fn = None
        self._x = None
        for k, v in kw.items():
            try:
                setattr(self, k, v)
            except AttributeError:
                pass

    @property
    def data(self):
        return self._x

    @data.setter
    def data(self, value):
        self._x = np.sort(value)
        self.n = len(value)

    @property
    def fn(self):
        if self._fn is not None:
            return lambda x: self._fn(x, *self._args, **self._kwargs)
        else:
            return None

    @fn.setter
    def fn(self, f):
        self._fn = f

    def set_args(self, *a, **kw):
        self._args = a
        self._kwargs = kw

    def _f_drop(self, i):
        from numpy import zeros_like

        mask = zeros_like(self.data)
        mask[i] = 1
        mx = ma.array(self.data, mask=mask)
        return np.array(self.fn(mx.compressed()))

    def pooled(self):
        assert self.data is not None and self.fn is not None
        return self.fn(self.data)

    def std(self):
        from numpy import sqrt, var, empty, searchsorted

        true_stat = self.fn(self.data)
        drop_stats = empty((self.n, len(true_stat)))

        # this part is specific for CI
        idx_lo, idx_hi = searchsorted(self.data, true_stat)
        self._fill_interval(drop_stats, (0, idx_lo), true_stat)
        self._fill_interval(drop_stats, (idx_lo, idx_hi), true_stat)
        self._fill_interval(drop_stats, (idx_hi, self.n - 1), true_stat)

        return sqrt((self.n - 1) * var(drop_stats, axis=0, ddof=1))

    def _fill_interval(self, drop_stats, interval, ev, splits=None):
        lo, hi = interval
        v_hi = self._f_drop(hi)
        v_lo = self._f_drop(lo)
        a = np.array([v_lo, v_hi, ev]).T
        g = lambda ar: (ar[0] == ar).all()
        if np.apply_along_axis(g, 1, a).all():
            for i in range(lo, hi + 1):
                drop_stats[i] = v_hi
        elif hi == lo:
            drop_stats[hi] = v_hi
        elif hi == lo + 1:
            drop_stats[hi] = v_hi
            drop_stats[lo] = v_lo
        else:
            splits = max(int(np.ceil(np.log10(hi - lo))), 2)
            dx = (hi - lo) // splits
            intervals = [[lo + dx * i, lo + dx * (i + 1)] for i in range(splits)]
            intervals[-1][1] = hi
            for inter in intervals:
                self._fill_interval(drop_stats, inter, np.mean((v_lo, v_hi), axis=0))


def n_dig(sx, sd):
    """
    Converts standard deviation to # of significand digits.

    *sx* is standard deviation of the statistic
    *sd* is standard deviation of the sample

    Adapted from section 7.9.2 of gum suppliment 1
    """
    l = np.log10(4 * sx)
    return np.floor(np.log10(sd)) - l + 1  # return as float or round now?



# This program is in the public domain
# Author: Paul Kienzle
"""
Termination conditions for solvers.

In order to decide when to stop fitting, the user needs to specify
stop conditions based on the optimizer history.  The test can use
the most recent value, the last n values or the entire computation.
See :mod:`monitor` for details.

Conditions can be composed, creating complicated criterion for
termination.  We may want to stop when we have found a minimum
and know its location.  We may be content just knowing the
minimum, and not worrying about the uncertainty in its location.
For some searches we want to be sure we that we are examining a
broad range of search space.  Here are some examples::

    import mystic.termination as stop

    # Stop when we know the location of the minimum; fail if run too long
    success = stop.Dx(0.001) & stop.Df(5)
    failure = stop.Steps(100)


    # Stop when we know the value of the minimum; fail if run too long
    success = stop.Dx(0.001) | stop.Df(5)
    failure = stop.Steps(100)

    # GA may want to run for a while, but only with a diverse population
    success = stop.Df(0.001,n=5) & stop.Steps(15)
    failure = stop.Steps(100) | stop.Dx(0.001)

When testing a conditional expression, a list of all conditions which
match is returned, or [] if no conditions match.


Predefined Conditions
---------------------

Dx: difference in x for step size test

    || (x_k - x_{k-1})/scale || < tolerance

Df: difference in f for improvement rate test

    | (f_k - f_{k-1})/scale | < tolerance

Rx: range in x for population distribution test

    max || (y - <y>)/ scale || < tolerance
    for y in population

Rf: range in f for value distribution test

    (max f(y) - min f(y))/scale < tolerance
    for y in population

Cx: constant x for target point

    ||(x_k - Z)/scale|| < tolerance

Cf: constant f for target value

    |f_k - A|/scale < tolerance

Steps: specific number of iterations

    k >= steps

Calls: specific number of function calls

    n >= calls

Time: wall clock time

    t_k >= time

CPU: CPU time

    t(CPU)_k >= time

Worse: fit is diverging

    (f_k - f_{k-1})/scale < -tolerance

Grad: fit is flat

    || del f_k || < tolerance

Feasible: value is in the feasible region ** Not implemented **

    f_k satisfies soft constraints

Invalid: values are not well defined ** Not implemented **

    isinf(y) or isinf(f(y)) or isnan(y) or isnan(f(y))
    for y in population


Distances and scaling
=====================

The following distance functions are predefined:

    norm_p(p): (sum |x_i|^p )^(1/p)  (generalized p-norm)
    norm_1:    sum |x_i|             (Manahattan distance)
    norm_2:    sqrt sum |x_i|^2      (Euclidian distance)
    norm_inf:  max |x_i|             (Chebychev distance)
    norm_min:  min |x_i|             (not a true norm)


The predefined scale factors in essence test for
percentage changes rather than absolute changes.
"""

import math

import numpy as np
from numpy import inf, isinf

from .condition import Condition


# ==== Norms ====
def norm_1(x):
    """1-norm: sum(|x_i|)"""
    return np.sum(abs(x))


def norm_2(x):
    """2-norm: sqrt(sum(|x_i|^2))"""
    return math.sqrt(np.sum(abs(x) ** 2))


def norm_inf(x):
    """inf-norm: max(|x_i|)"""
    return max(abs(x))


def norm_min(x):
    """min-norm: min(|x_i|); this is not a true norm"""
    return min(abs(x))


def norm_p(p):
    """p-norm: sum(|x_i|^p)^(1/p)"""
    if isinf(p):
        if p < 0:
            return norm_min
        else:
            return norm_inf
    elif p == 1:
        return norm_1
    elif p == 2:
        return norm_2
    else:
        return lambda x: np.sum(abs(x) ** p) ** (1 / p)


# ==== Conditions ====
class Dx(Condition):
    """
    Improvement in x.

    This condition measures the improvement over the last n iterations
    in terms of how much the value of x has changed::

        norm((x[k]- x[k-n])/scale) < tol

    where x[k] is the best parameter set for iteration step k.

    The scale factor to use if scaled is upper bound - lower bound
    if the parameter is bounded, or 1/2 (|x[k]| + |x[k-n]|)
    if the parameter is unbounded, with protection against a scale
    factor of zero.

    Parameters::

        *tol* (float = 0.001)
            tolerance to test against
        *norm* ( f(vector): float  =  norm_2)
            norm to use to measure the size of x.  Predefined norms
            include norm_1, norm_2, norm_info, norm_min and norm_p(p)
        *n* (int = 1)
            number of steps back in history to compare
        *scaled* (boolean = True)
            whether to use raw or scaled differences in the norm

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, tol=0.001, norm=norm_2, n=1, scaled=True):
        self.tol = tol
        self.norm = norm
        self.n = n
        self.scaled = scaled

    def _scaled_condition(self, history):
        x1, x2 = history.point[0], history.point[self.n]
        scale = history.upper_bound - history.lower_bound
        scale[isinf(scale)] = ((abs(x1) + abs(x2)) / 2)[isinf(scale)]
        scale[scale == 0] = 1
        return self.norm((x2 - x1) / scale)

    def _raw_condition(self, history):
        x1, x2 = history.point[0], history.point[self.n]
        return self.norm(x2 - x1)

    def config_history(self, history):
        """
        Needs the previous n points from history.
        """
        if self.tol > 0:
            history.requires(point=self.n + 1)

    def _subcall(self, history):
        """
        Returns True if the tolerance is met.
        """
        if self.tol == 0 or len(history.point) < self.n + 1:
            return False  # Cannot succeed until at least n generations
        elif self.scaled:
            return self._scaled_condition(history)
        else:
            return self._raw_condition(history)

    def __call__(self, history):
        return self._subcall(history) < self.tol

    def completeness(self, history):
        return self._subcall(history) / self.tol

    def __str__(self):
        if self.scaled:
            return "||(x[k] - x[k-%d])/range|| < %g" % (self.n, self.tol)
        else:
            return "||x[k] - x[k-%d]|| < %g" % (self.n, self.tol)


class Df(Condition):
    """
    Improvement in F(x)

    This condition measures the improvement over the last n iterations
    in terms of how much the value of the function has changed::

        | (F[k] - F[k-1])/scale | < tol

    where F[k] is the value for the best parameter set for iteration step k.

    The scale factor to use is 1/2 (|F(k)| + |F(k-n)|) with protection
    against zero.

    Parameters::

        *tol* (float = 0.001)
            tolerance to test against
        *n* (int = 1)
            number of steps back in history to compare
        *scaled* (boolean = True)
            whether to use raw or scaled differences in the norm

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, tol=0.001, n=1, scaled=True):
        self.tol = tol
        self.n = n
        self.scaled = scaled

    def _scaled_condition(self, history):
        f1, f2 = history.value[0], history.value[self.n]
        scale = (abs(f1) + abs(f2)) / 2
        if scale == 0:
            scale = 1
        # print "Df",f1,f2,abs(float(f2-f1)/scale),self.tol
        return abs(float(f2 - f1) / scale)

    def _raw_condition(self, history):
        f1, f2 = history.value[0], history.value[self.n]
        return abs(f2 - f1)

    def config_history(self, history):
        """
        Needs the previous n points from history.
        """
        if self.tol > 0:
            history.requires(value=self.n + 1)

    def __call__(self, history):
        """
        Returns True if the tolerance is met.
        """
        if self.tol == 0 or len(history.value) < self.n + 1:
            return False  # Cannot succeed until at least n generations
        elif self.scaled:
            return self._scaled_condition(history) < self.tol
        else:
            return self._raw_condition(history) < self.tol

    def __str__(self):
        if self.scaled:
            return "|F[k]-F[k-%d]| / (|F[k]|+|F[k-%d]|)/2 < %g" % (self.n, self.n, self.tol)
        else:
            return "|F[k]-F[k-%d]| < %g" % (self.n, self.tol)


class Worse(Condition):
    """
    Worsening of F(x)

    This condition measures whether the fit is diverging.  You may want
    to use this for non-greedy optimizers which can get worse over time::

        (F[k] - F[k-1])/scale < -tol

    where F[k] is the value for the best parameter set for iteration step k.

    The scale factor to use is 1/2 (|F(k)| + |F(k-n)|) with protection
    against zero.

    Parameters::

        *tol* (float = 0)
            tolerance to test against
        *n* (int = 1)
            number of steps back in history to compare
        *scaled* (boolean = True)
            whether to use raw or scaled differences in the norm

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, tol=0, n=1, scaled=True):
        self.tol = tol
        self.n = n
        self.scaled = scaled

    def _scaled_condition(self, history):
        f1, f2 = history.value[0], history.value[self.n]
        scale = (abs(f1) + abs(f2)) / 2
        if scale == 0:
            scale = 1
        return float(f2 - f1) / scale

    def _raw_condition(self, history):
        f1, f2 = history.value[0], history.value[self.n]
        return f2 - f1

    def config_history(self, history):
        """
        Needs the previous n points from history.
        """
        history.requires(value=self.n + 1)

    def __call__(self, history):
        """
        Returns True if the tolerance is met.
        """
        if self.tol == 0 or len(history.value) < self.n + 1:
            return False  # Cannot succeed until at least n generations
        elif self.scaled:
            return self._scaled_condition(history) < -self.tol
        else:
            return self._raw_condition(history) < -self.tol

    def __str__(self):
        if self.scaled:
            return "F[k]-F[k-%d] / (|F[k]|+|F[k-%d]|)/2 < -%g" % (self.n, self.n, self.tol)
        else:
            return "F[k]-F[k-%d] < -%g" % (self.n, self.tol)


class Grad:
    """
    Flat function value

    This condition measures whether the fit surface is flat near the best
    value.  This only works for fits which compute the gradient.

        || del F[k]/scale || < tol

    where F[k] is the value for the best parameter set for iteration step k.

    The scale factor to use is |F(k)| with protection against zero.

    Parameters::

        *tol* (float = 0.001)
            tolerance to test against
        *norm* ( f(vector): float  =  norm_2)
            norm to use to measure the size of x.  Predefined norms
            include norm_1, norm_2, norm_info, norm_min and norm_p(p)
        *scaled* (boolean = True)
            whether to use raw or scaled differences in the norm

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, tol=0.001, norm=norm_2, scaled=True):
        self.tol = tol
        self.norm = norm
        self.scaled = scaled

    def _scaled_condition(self, history):
        df = history.gradient[0]
        f = history.value[0]
        scale = abs(f)
        if scale == 0:
            scale = 1
        return self.norm(df / float(scale))

    def _raw_condition(self, history):
        df = history.gradient[0]
        return self.norm(df)

    def config_history(self, history):
        """
        Needs the previous n points from history.
        """
        if self.tol > 0:
            history.requires(gradient=1, value=1)

    def __call__(self, history):
        """
        Returns True if the tolerance is met.
        """
        if self.tol == 0 or len(history.gradient) < 1:
            return False  # Cannot succeed until at least n generations
        elif self.scaled:
            return self._scaled_condition(history) < self.tol
        else:
            return self._raw_condition(history) < self.tol

    def __str__(self):
        if self.scaled:
            return "|| del F[k]/F[k] || < %g" % (self.tol)
        else:
            return "|| del F[k] || < %g" % (self.tol)


class r_best:
    """
    Measure of population radius based on distance from the best.

        max ||(y - x[k])/scale|| for y in population

    scipy.optimize.fmin uses r_best(norm_inf) as its measure of radius.
    """

    def __init__(self, norm):
        self.norm = norm

    def __call__(self, population, best, scale):
        P = np.asarray(population)
        r = max(self.norm(p - best) / scale for p in P)
        return r


class r_centroid:
    """
    Measure of population radius based on distance from the centroid.

        max ||(y - <y>)/scale|| for y in population
    """

    def __init__(self, norm):
        self.norm = norm

    def __call__(self, population, best, scale):
        P = np.asarray(population)
        c_i = np.mean(P, axis=0)
        r = max(self.norm(p - c_i) / scale for p in P)
        return r


def r_boundingbox(population, best, scale):
    """
    Measure of population radius based on the volume of the bounding box.

        (product (max(y_i) - min(y_i))/scale)**1/k  for i in dimensions-k
    """
    P = np.asarray(population)
    lo = max(P, index=0)
    hi = max(P, index=0)
    r = np.prod((hi - lo) / scale) ** (1 / len(hi))
    return r


class r_hull:
    """
    Measure of population radius based on maximum diameter in convex hull.

        1/2 max || (y1 - y2)/scale || for y1,y2 in population
    """

    def __init__(self, norm):
        self.norm = norm

    def __call__(self, population, best, scale):
        r = 0
        for i, y1 in enumerate(population):
            for y2 in population[i + 1 :]:
                d = self.norm(y2 - y1) / scale
                if d > r:
                    r = d
        return r / 2


class Rx(Condition):
    """
    Domain size

    This condition measures the size of the population domain.  Some
    algorithms are done when the domain size shrinks while others have
    failed if the domain size shrinks.

    There are a number of ways of measuring the domain size::

        r_best(norm) : radius from best point

            max ||(y - x[k])/scale|| for y in population

        r_centroid(norm) : radius from centroid

            max ||(y - <y>)/scale|| for y in population

        r_boundingbox : radius from bounding box

            (product (max(y_i) - min(y_i))/scale)**1/k  for i in dimensions-k

        r_hull(norm) : radius from convex hull

            1/2 max || (y1 - y2)/scale || for y1,y2 in population

    scale is determined from the fit bounds (max-min) or the
    values sum(|y_i|)/n, with protection against zero values.

    Parameters::

        *tol* (float = 0.001)
            tolerance to test against
        *radius* (function(history,best,scale): float = r_centroid(norm_2))
            measure of domain size
        *scaled* (boolean = True)
            whether to use raw or scaled differences in the norm

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, tol=0, radius=r_centroid(norm_2), scaled=False):
        self.tol = tol
        self.radius = radius
        self.scaled = scaled

    def _scaled_condition(self, history):
        P = np.asarray(history.population_points[0])
        scale = history.upper_bound - history.lower_bound
        idx = isinf(scale)
        if any(idx):
            range = np.sum(abs(P), axis=0) / P.shape[0]
            scale[idx] = range[idx]
        scale[scale == 0] = 1
        r = self.radius(P, history.point[0], scale)
        # print "Rx=%g, scale=%g"%(r,scale)
        return r

    def _raw_condition(self, history):
        P = np.asarray(history.population_points[0])
        r = self.radius(P, history.point[0], scale=1.0)
        # print "Rx=%g"%r
        return r

    def config_history(self, history):
        """
        Needs the previous n points from history.
        """
        if self.tol > 0:
            history.requires(population_points=1, point=1)

    def __call__(self, history):
        """
        Returns True if the tolerance is met.
        """
        if self.tol == 0 or len(history.population_points) < 1:
            return False
        elif self.scaled:
            return self._scaled_condition(history) < self.tol
        else:
            return self._raw_condition(history) < self.tol

    def __str__(self):
        if self.scaled:
            return "radius(population/scale) < %g" % (self.tol)
        else:
            return "radius(population) < %g" % (self.tol)


class Rf(Condition):
    """
    Range size

    This condition measures the size of the population range::

        (max f(y) - min f(y))/scale < tol

    for y in the current population
    scale is mean(|f(y)|)

    Parameters::

        *tol* (float = 0.001)
            tolerance to test against
        *scaled* (boolean = True)
            whether to use raw or scaled differences in the norm

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, tol=0, scaled=True):
        self.tol = tol
        self.scaled = scaled

    def _scaled_condition(self, history):
        Pf = np.asarray(history.population_values)
        scale = np.mean(abs(Pf))
        if scale == 0:
            scale = 1
        r = float(np.max(Pf) - np.min(Pf)) / scale
        # print "Rf = %g, scale=%g"%(r,scale)
        return r

    def _raw_condition(self, history):
        P = np.asarray(history.population_values)
        r = np.max(P) - np.min(P)
        # print "Rf = %g"%r
        return r

    def config_history(self, history):
        """
        Needs the previous n points from history.
        """
        if self.tol > 0:
            history.requires(population_values=1)

    def __call__(self, history):
        """
        Returns True if the tolerance is met.
        """
        if self.tol == 0 or len(history.population_values) < 1:
            return False
        elif self.scaled:
            return self._scaled_condition(history) < self.tol
        else:
            return self._raw_condition(history) < self.tol

    def __str__(self):
        if self.scaled:
            return "(max(F(p)) - min(F(p))/mean(|F(p)|) < %g" % (self.tol)
        else:
            return "max(F(p)) - min(F(p)) < %g" % (self.tol)


class Cx(Condition):
    """
    Target point

    This condition measures the distance from the best point
    to some target point::

       ||(x_k - Z)/scale|| < tol

    scale is fit range if given,  |Z_i|, or 1 if Z_i=0

    Paramaters::

        *tol* (float = 0.001)
            tolerance to test against
        *point* (array = 0)
            target point
        *norm* ( f(vector): float  =  norm_2)
            norm to use to measure the size of x.  Predefined norms
            include norm_1, norm_2, norm_info, norm_min and norm_p(p)
        *scaled* (boolean = True)
            whether to use raw or scaled differences in the norm

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, tol=0.001, point=0, norm=norm_2, scaled=True):
        self.tol = tol
        self.point = point
        self.norm = norm_2
        self.scaled = scaled

    def _scaled_condition(self, history):
        x = history.point[0]
        scale = history.upper_bound - history.lower_bound
        scale[isinf(scale)] = abs(self.point)[isinf(scale)]
        scale[scale == 0] = 1
        return self.norm((x - self.point) / scale)

    def _raw_condition(self, history):
        x = history.point[0]
        return self.norm(x - self.point)

    def config_history(self, history):
        """
        Needs the previous point from history.
        """
        if self.tol > 0:
            history.requires(point=1)

    def __call__(self, history):
        """
        Returns True if the tolerance is met.
        """
        if self.tol == 0 or len(history.point) < 1:
            return False
        elif self.scaled:
            return self._scaled_condition(history) < self.tol
        else:
            return self._raw_condition(history) < self.tol

    def __str__(self):
        if self.scaled:
            return "||(x[k] - Z)/range|| < %g" % (self.tol)
        else:
            return "||x[k] - Z|| < %g" % (self.tol)


class Cf(Condition):
    """
    Target value

    This condition measures the distance from the best value
    to some target value::

       |(f_k - A)/scale| < tol

    scale is |A| or 1 if A=0

    Paramaters::

        *tol* (float = 0.001)
            tolerance to test against
        *value* (float = 0)
            target value
        *scaled* (boolean = True)
            whether to use raw or scaled differences in the norm

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, tol=0.001, value=0, scaled=True):
        self.tol = tol
        self.value = value
        if value == 0:
            scaled = False
        self.scaled = scaled

    def _scaled_condition(self, history):
        value = history.value[0]
        return abs(float(value - self.value) / self.value)

    def _raw_condition(self, history):
        value = history.value[0]
        return abs(value - self.value)

    def config_history(self, history):
        """
        Needs the previous point from history.
        """
        if self.tol > 0:
            history.requires(value=1)

    def __call__(self, history):
        """
        Returns True if the tolerance is met.
        """
        if self.tol == 0 or len(history.value) < 1:
            return False
        elif self.scaled:
            return self._scaled_condition(history) < self.tol
        else:
            return self._raw_condition(history) < self.tol

    def __str__(self):
        if self.scaled:
            return "|(F[k] - A)/A| < %g" % (self.tol)
        else:
            return "|F[k] - A| < %g" % (self.tol)


class Steps(Condition):
    """
    Specific number of iterations

    This condition test the number of iterations of a fit::

        k >= steps

    Parameters::

        *steps* int (1000)
            total number of steps

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, steps=np.inf):
        self.steps = steps

    def __call__(self, history):
        if len(history.step) < 1:
            return False
        return history.step[0] >= self.steps

    def config_history(self, history):
        history.requires(step=1)

    def __str__(self):
        return "steps >= %d" % self.steps


class Calls(Condition):
    """
    Specific number of function calls

    This condition tests the number of function evaluations::

        n_k >= calls

    Parameters::

        *calls* int (inf)
            total number of function calls

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, calls=np.inf):
        self.calls = calls

    def __call__(self, history):
        if len(history.calls) < 1:
            return False
        return history.calls[0] >= self.calls

    def config_history(self, history):
        history.requires(calls=1)

    def __str__(self):
        return "calls >= %d" % self.calls


class Time(Condition):
    """
    Wall clock time.

    This condition tests wall clock time::

        t_k >= time

    Parameters::

        *time* float (inf)
            Time since start of job in seconds

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, time=inf):
        self.time = time

    def __call__(self, history):
        if len(history.time) < 1:
            return False
        return history.time[0] >= self.time

    def config_history(self, history):
        history.requires(time=1)

    def __str__(self):
        return "time >= %g" % self.time


class CPU(Condition):
    """
    CPU time.

    This condition tests CPU time::

        t(CPU)_k >= time

    Parameters::

        *time* float (inf)
            time since start of job in seconds

    Returns::

        *condition* (f(history) : boolean)
            a callable returning true if the condition is met
    """

    def __init__(self, time=np.inf):
        self.time = time

    def __call__(self, history):
        if len(history.cpu_time) < 1:
            return False
        return history.cpu_time[0] >= self.time

    def config_history(self, history):
        history.requires(cpu_time=1)

    def __str__(self):
        return "cpu_time >= %g" % self.time


"""
class Feasible: value can be used ** Not implemented **

    f_k satisfies soft constraints

class Invalid: values are well defined

    isinf(y) or isinf(f(y)) or isnan(y) or isnan(f(y))

    for y in population
"""


def parse_condition(cond):
    import math
    from . import stop

    return eval(cond, stop.__dict__.copy().update(math.__dict__))



"""
Population initialization routines.

To start the analysis an initial population is required.  This will be
an array of size M x N, where M is the number of dimensions in the fitting
problem and N is the number of Markov chains.

Two functions are provided:

1. lhs_init(N, bounds) returns a latin hypercube sampling, which tests every
parameter at each of N levels.

2. cov_init(N, x, cov) returns a Gaussian sample along the ellipse
defined by the covariance matrix, cov.  Covariance defaults to
diag(dx) if dx is provided as a parameter, or to I if it is not.

Additional options are random box: rand(M, N) or random scatter: randn(M, N).
"""

__all__ = ["lhs_init", "cov_init"]

from numpy import asarray, diag, empty, eye

from . import util


def lhs_init(N, bounds):
    """
    Latin Hypercube Sampling

    Returns an array whose columns each have *N* samples from equally spaced
    bins between *bounds=(xmin, xmax)* for the column.  DREAM bounds
    objects, with bounds.low and bounds.high can be used as well.

    Note: Indefinite ranges are not supported.
    """
    try:
        xmin, xmax = bounds.low, bounds.high
    except AttributeError:
        xmin, xmax = bounds

    # Define the size of xmin
    nvar = len(xmin)
    # Initialize array ran with random numbers
    ran = util.rng.rand(N, nvar)

    # Initialize array s with zeros
    s = empty((N, nvar))

    # Now fill s
    for j in range(nvar):
        # Random permutation
        idx = util.rng.permutation(N) + 1
        p = (idx - ran[:, j]) / N
        s[:, j] = xmin[j] + p * (xmax[j] - xmin[j])

    return s


def cov_init(N, x, cov=None, dx=None):
    """
    Initialize *N* sets of random variables from a gaussian model.

    The center is at *x* with an uncertainty ellipse specified by the
    1-sigma independent uncertainty values *dx* or the full covariance
    matrix uncertainty *cov*.

    For example, create an initial population for 20 sequences for a
    model with local minimum x with covariance matrix C::

        pop = cov_init(cov=C, x=x, N=20)
    """
    # return mean + dot(util.rng.randn(N, len(mean)), chol(cov))
    if cov is None and dx is None:
        cov = eye(len(x))
    elif cov is None:
        cov = diag(asarray(dx) ** 2)
    return util.rng.multivariate_normal(mean=x, cov=cov, size=N)


def demo():
    from numpy import arange

    print("Three ways of calling cov_init:")
    print("with cov", cov_init(N=4, x=[5, 6], cov=diag([0.1, 0.001])))
    print("with dx", cov_init(N=4, x=[5, 6], dx=[0.1, 0.001]))
    print("with nothing", cov_init(N=4, x=[5, 6]))
    print("""
The following array should have four columns.  Column 1 should have the
numbers from 10 to 19, column 2 from 20 to 29, etc.  The columns are in
random order with a random fractional part.
""")
    pop = lhs_init(N=10, bounds=(arange(1, 5), arange(2, 6))) * 10
    print(pop)


if __name__ == "__main__":
    demo()



"""
Pylab plotting utilities.
"""

__all__ = ["auto_shift", "coordinated_colors", "dhsv", "next_color", "plot_quantiles", "form_quantiles"]


def auto_shift(offset):
    """
    Return a y-offset coordinate transform for the current axes.

    Each call to auto_shift increases the y-offset for the next line by
    the given number of points (with 72 points per inch).

    Example::

        from matplotlib import pyplot as plt
        from bumps.plotutil import auto_shift
        trans = auto_shift(plt.gca())
        plot(x, y, trans=trans)
    """
    from matplotlib.transforms import ScaledTranslation
    import pylab

    ax = pylab.gca()
    if ax.lines and hasattr(ax, "_auto_shift"):
        ax._auto_shift += offset
    else:
        ax._auto_shift = 0
    trans = pylab.gca().transData
    if ax._auto_shift:
        trans += ScaledTranslation(0, ax._auto_shift / 72.0, pylab.gcf().dpi_scale_trans)
    return trans


# ======== Color functions ========


def next_color(axes=None):
    """
    Return the next color in the plot color cycle.

    Example::

        from matplotlib import pyplot as plt
        from bumps.plotutil import next_color, dhsv
        color = next_color()
        plt.errorbar(x, y, yerr=dy, fmt='.', color=color)
        # Draw the theory line with the same color as the data, but darker
        plt.plot(x, y, '-', color=dhsv(color, dv=-0.2))
    """
    import pylab

    if axes is None:
        axes = pylab.gca()
    lines = axes._get_lines
    try:
        base = lines.get_next_color()
    except Exception:
        try:  # Cruft 1.7 - 3.7?
            base = next(lines.prop_cycler)["color"]
        except Exception:
            try:  # Cruft 1.4-1.6?
                base = next(lines.color_cycle)
            except Exception:  # Cruft 1.3 and earlier
                base = lines._get_next_cycle_color()
    return base


def coordinated_colors(base=None):
    """
    Return a set of coordinated colors as c['base|light|dark'].

    If *base* is not provided, use the next color in the color cycle as
    the base.  Light is bright and pale, dark is dull and saturated.
    """
    if base is None:
        base = next_color()
    return dict(
        base=base,
        light=dhsv(base, dv=+0.3, ds=-0.2),
        dark=dhsv(base, dv=-0.25, ds=+0.35),
    )


def dhsv(color, dh=0.0, ds=0.0, dv=0.0, da=0.0):
    """
    Modify color on hsv scale.

    *dv* change intensity, e.g., +0.1 to brighten, -0.1 to darken.
    *dh* change hue
    *ds* change saturation
    *da* change transparency

    Color can be any valid matplotlib color.  The hsv scale is [0,1] in
    each dimension.  Saturation, value and alpha scales are clipped to [0,1]
    after changing.  The hue scale wraps between red to violet.

    :Example:

    Make sea green 10% darker:

        >>> from bumps.plotutil import dhsv
        >>> darker = dhsv('seagreen', dv=-0.1)
        >>> print([int(v*255) for v in darker])
        [37, 113, 71, 255]
    """
    from matplotlib.colors import colorConverter
    from colorsys import rgb_to_hsv, hsv_to_rgb
    from numpy import clip, array, fmod

    r, g, b, a = colorConverter.to_rgba(color)
    # print "from color",r,g,b,a
    h, s, v = rgb_to_hsv(r, g, b)
    s, v, a = [clip(val, 0.0, 1.0) for val in (s + ds, v + dv, a + da)]
    h = fmod(h + dh, 1.0)
    r, g, b = hsv_to_rgb(h, s, v)
    # print "to color",r,g,b,a
    return array((r, g, b, a))


# ==== Specialized plotters =====


def plot_quantiles(x, y, contours, color, alpha=None, axes=None):
    """
    Plot quantile curves for a set of lines.

    *x* is the x coordinates for all lines.

    *y* is the y coordinates, one row for each line.

    *contours* is a list of confidence intervals expressed as percents.

    *color* is the color to use for the quantiles.  Quantiles are draw as
    a filled region with alpha transparency.  Higher probability regions
    will be covered with multiple contours, which will make them lighter
    and more saturated.

    *alpha* is the transparency level to use for all fill regions.  The
    default value, alpha=2./(#contours+1), works pretty well.
    """
    _, q = form_quantiles(y, contours)
    _plot_quantiles(x, q, color, alpha, axes=axes)


def _plot_quantiles(x, q, color, alpha, axes=None):
    import pylab

    # print "p",p
    # print "q",q[:,:,0]
    # print "y",y[:,0]
    if axes is None:
        axes = pylab.gca()
    if alpha is None:
        alpha = 2.0 / (len(q) + 1)
    edgecolor = dhsv(color, ds=-(1 - alpha), dv=(1 - alpha))
    for lo, hi in q:
        axes.fill_between(x, lo, hi, facecolor=color, edgecolor=edgecolor, alpha=alpha)


def form_quantiles(y, contours):
    """
    Return quantiles and values for a list of confidence intervals.

    *contours* is a list of confidence interfaces [a, b,...] expressed as
    percents.

    Returns:

    *quantiles* is a list of intervals [[a_low, a_high], [b_low, b_high], ...]
    in [0,1].

    *values* is a list of intervals [[A_low, A_high], ...] with one entry in
    A for each row in y.
    """
    from numpy import reshape
    from scipy.stats.mstats import mquantiles

    p = _convert_contours_to_probabilities(reversed(sorted(contours)))
    q = mquantiles(y, prob=p, axis=0)
    p = reshape(p, (2, -1))
    q = reshape(q, (-1, 2, len(y[0])))
    return p, q


def _convert_contours_to_probabilities(contours):
    """
    Given confidence intervals [a, b,...] as percents, return probability
    in [0,1] for each interval as [a_low, a_high, b_low, b_high, ...].
    """
    from numpy import hstack

    # lower quantile for ci in percent = (100 - ci)/2
    # upper quantile = 100 - lower quantile = 100 - (100-ci)/2 = (100 + ci)/2
    # divide by an additional 100 to get proportion from 0 to 1
    return hstack([(100.0 - p, 100.0 + p) for p in contours]) / 200.0



"""
Generic minimizers

The general optimization algorithm is as follows::

    fit = Minimizer(problem=Problem(),
                 strategy=Strategy(),
                 monitors=[],
                 success=Df(1e-5) & Dx(1e-5),
                 failure=Calls(10000) | Steps(100))
    population = fit.start()
    while True:
        result = list(map(fit.problem, population))
        fit.update(population, result)
        if fit.isdone(): break
        population = fit.step()

    if fit.successful:
        print "Converged"
    else:
        print "Stopped with",", ".join(str(c) for c in fit.failure_cond)

Variations are possible, such as the multistep process where the
program first generates an initial population, then each time it
is run generates an updated population based on the results of
submitting the previous population to a batch queue.

History traces
==============

Stopping conditions use history traces to evaluate whether the program
should continue.  When adding new stopping conditions, the programmer
must define a config_history method to indicate which values are needed.
For example::

    def config_history(self, history):
        history.requires(points=3)

When checking convergence, the programmer must still check that enough
history is available for the test. For example,
::

    def __call__(self, history):
        from numpy.linalg import norm
        if len(history.value) < 3: return False
        return norm(history.value[0] - history.value[2]) < 0.001

The optimizer will make sure that config_history is called for each
condition before the fit starts::

    for t in success.primitives()|failure.primitives():
        t.config_history()

Each optimizer can control which values it wants to monitor.  For
consistency and ease of use, monitor names should be chosen from
the standard names below.  As new optimizers are created, the
list of standard names may expand.

Fixed properties of history::

    *ndim* (int)
         problem dimension
    *lower_bound*, *upper_bound* ([float])
         bounds constraints

Accumulated properties::

    *step* (int)
         iteration number
    *calls* (int)
        cumulative number of function evaluations
    *time* (seconds)
        cumulative wall clock time
    *cpu_time* (seconds)
        cumulative CPU time
    *value* (float)
        best function value
    *point* (vector)
        parameter values for the best function value
    *gradient* (vector)
        del f: gradient at best point (if available)
    *hessian* (matrix)
        del**2 F: Hessian at best point (if available)
    *population_values* (vector)
        function values of the current population (if available)
    *population_points* (matrix)
        parameter values for the current population (if available)
"""

# TODO: Ordered fits
#
# Want a list of strategies and a parameter subset associated with each
# strategy.  The fit is nested, with the outermost parameters are set
# to a particular value, then the inner parameters are optimized for
# those values before an alternative set of outer parameters is considered.
#
# Will probably want the inner strategy to be some sort of basin hopping
# method with the new value jumping around based on the best value for
# those parameters.  This is complicated because the first fit should be
# done using a more global method.
#
# In the context of simultaneous fitting, it would be natural to partition
# the parameters so that the dependent parameters of the slowest models
# are varied last.  This can be done automatically if each model has
# a cost estimate associated with it.
#
# TODO: Surrogate models
#
# Want to use surrogate models for the expensive models when far from the
# minimum, only calculating the real model when near the minimum.

import os
import time

import numpy as np


def default_mapper(f, x):
    return list(map(f, x))


def cpu_time():
    """Current cpu time for this process"""
    user_time, sys_time, _, _, _ = os.times()
    return user_time + sys_time


class Minimizer:
    """
    Perform a minimization.
    """

    def __init__(self, problem=None, strategy=None, monitors=[], history=None, success=None, failure=None):
        self.problem = problem
        self.strategy = strategy
        # Ask strategy to fill in the default termination conditions
        # in case the user doesn't supply them.
        defsucc, deffail = strategy.default_termination_conditions(problem)
        self.success = success if success is not None else defsucc
        self.failure = failure if failure is not None else deffail
        self.monitors = monitors
        self.history = history
        self.reset()

    def minimize(self, mapper=default_mapper, abort_test=None, resume=False):
        """
        Run the solver to completion, returning the best point.

        Note: only used stand-alone, not within fit service
        """
        self.time = time.time()
        self.remote_time = -cpu_time()
        population = self.step() if resume else self.start()
        try:
            while True:
                result = mapper(self.problem, population)
                # print "map result",result
                self.update(population, result)
                # print self.history.step, self.history.value
                if self.isdone():
                    break  # STOPHERE combine
                if abort_test is not None and abort_test():
                    break
                population = self.step()
        except KeyboardInterrupt:
            pass
        return self.history.point[0]

    __call__ = minimize

    def reset(self):
        """
        Clear the solver history.
        """
        self.history.clear()
        self.history.provides(
            calls=1, time=1, cpu_time=1, step=1, point=1, value=1, population_points=0, population_values=0
        )
        for c in self.success.primitives() | self.failure.primitives():
            c.config_history(self.history)
        for m in self.monitors:
            m.config_history(self.history)
        self.strategy.config_history(self.history)

    def start(self):
        """
        Start the optimization but generating an initial population.
        """
        # Reset the timers so we know how long the fit takes.
        # We are cheating by initializing remote_time to -cpu_time, then
        # later adding cpu_time back to remote_time to get the total cost
        # of local and remote computation.

        if len(self.problem.getp()) == 0:
            raise ValueError("Problem has no fittable parameters")

        return self.strategy.start(self.problem)

    def update(self, points, values):
        """
        Collect statistics on time and resources
        """
        if hasattr(values, "__cpu_time__"):
            self.remote_time += values.__cpu_time__

        # Locate the best member of the population
        values = np.asarray(values)
        # print("values",values,file=sys.stderr)

        # Update the history
        self.history.update(
            time=time.time() - self.time,
            cpu_time=cpu_time() + self.remote_time,
            population_points=points,
            population_values=values,
        )
        self.history.accumulate(step=1, calls=len(points))

        self.strategy.update(self.history)

        minidx = np.argmin(values)
        self.history.update(
            point=points[minidx],
            value=values[minidx],
        )

        # Tell all the monitors that the history has been updated
        for m in self.monitors:
            m(self.history)

    def step(self):
        """
        Generate the next population to evaluate.
        """
        return self.strategy.step(self.history)

    def isdone(self):
        """
        Check if the fit has converged according to the criteria proposed
        by the user and/or the optimizer.

        Returns True if either the fit converged or is forced to stop for
        some other reason.

        Sets the following attributes::

            *successful* (boolean)
                True if the fit converged
            *success_cond* ([Condition])
                Reasons for success or lack of success
            *failed* (boolean)
                True if the fit should stop
            *failure_cond* ([Condition])
                Reasons for failure or lack of failure

        Note that success and failure can occur at the same time if
        for example the convergence criteria are met but the resource
        limits were exceeded.
        """
        self.successful, self.success_cond = self.success.status(self.history)
        self.failed, self.failure_cond = self.failure.status(self.history)
        return self.successful or self.failed

    def termination_condition(self):
        if self.successful:
            return "succeeded with " + ", ".join(str(c) for c in self.success_cond)
        else:
            return (
                "failed with "
                + ", ".join(str(c) for c in self.failure_cond)
                + " and "
                + ", ".join(str(c) for c in self.success_cond)
            )


class Strategy:
    """
    Optimization strategy to use.

    The doc string for the strategy will be used in the construction of
    the doc strings for the simple optimizer interfaces, with a description
    of the standard optimizer options at the end.   The following template
    works well::

        Name of optimization strategy

        Brief description of the strategy

        Optimizer parameters::

            *argument* is the description of the argument

        Additional details about the solver.

    The __init__ arguments for the strategy will be passed in through
    the simple optimizer interface.  Be sure to make them all keyword
    arguments.
    """

    def config_history(self, history):
        """
        Specify which information needs to be preserved in history.

        For example, parallel tempering needs to save its level values::

            history.requires(levels=2)
        """
        pass

    def update(self, history):
        """
        Update history with optimizer specific state information.

        Note: standard history items (step, calls, value, point,
        population_values, population_points, time, cpu_time) are
        already recorded.  Additional items should be recorded
        directly in the trace.  For example::

            history.levels.put([1,2,3])
        """
        pass

    def start(self, problem):
        """
        Generate the initial population.

        Returns a matrix *P* of points to be evaluated.
        """
        raise NotImplementedError

    def step(self, history):
        """
        Generate the next population.

        Returns a matrix *P* of points to be evaluated.

        *history* contains the previous history of the computation,
        including any fields placed by :meth:`update`.
        """
        raise NotImplementedError



"""
Bounds handling.

Use bounds(low, high, style) to create a bounds handling object.  This
function operates on a point x, transforming it so that all dimensions
are within the bounds.  Options are available, including reflecting,
wrapping, clipping or randomizing the point, or ignoring the bounds.

The returned bounds object should have an apply(x) method which
transforms the point *x*.
"""

__all__ = ["make_bounds_handler", "Bounds", "ReflectBounds", "ClipBounds", "FoldBounds", "RandomBounds", "IgnoreBounds"]

import numpy as np
from numpy import inf, isinf
from . import util
from .compiled import dll

try:
    from numba import njit
except ImportError:

    def njit(*args, **kw):
        return lambda f: f


def make_bounds_handler(bounds, style="reflect"):
    """
    Return a bounds object which can update the bounds.

    Bounds handling *style* name is one of::

        reflect:   reflect off the boundary
        clip:      stop at the boundary
        fold:      wrap values to the other side of the boundary
        randomize: move to a random point in the bounds
        none:      ignore the bounds

    With semi-infinite intervals folding and randomizing aren't well
    defined, and reflection is used instead.

    With finite intervals the the reflected or folded point may still be
    outside the bounds (which can happen if the step size is too large),
    and a random uniform value is used instead.
    """
    if bounds is None:
        return IgnoreBounds()

    low, high = bounds

    # Do boundary handling -- what to do when points fall outside bound
    style = style.lower()
    if style == "reflect":
        f = ReflectBounds(low, high)
    elif style == "clip":
        f = ClipBounds(low, high)
    elif style == "fold":
        f = FoldBounds(low, high)
    elif style == "randomize":
        f = RandomBounds(low, high)
    elif style == "none" or style is None:
        f = IgnoreBounds()
    else:
        raise ValueError("bounds style %s is not valid" % style)
    return f


class Bounds(object):
    """
    Base class for all times of bounds objects.
    """

    c_interface = None  # type: Callable[[int, int, Any, Any, Any], None]
    low = None  # type: np.ndarray
    high = None  # type: np.ndarray

    @staticmethod
    def apply(minn, maxn, pop):
        """Force pop (population) values within bounds"""
        raise NotImplementedError

    def __call__(self, population):
        """
        Force parameter values within bounds for each member of the population
        (population is expected to be a 2d array of shape (M, N) where
         - M is the size of the population
         - N is then number of parameters
        Returns population for convenience.  E.g., y = bounds(x+0)
        """
        if self.c_interface is not None:
            self.c_interface(len(population), len(self.low), population.ctypes, self.low.ctypes, self.high.ctypes)
        else:
            self.apply(self.low, self.high, population)
        return population


class ReflectBounds(Bounds):
    """
    Reflect parameter values into bounded region
    """

    c_interface = dll.bounds_reflect if dll else None

    def __init__(self, low, high):
        self.low, self.high = [np.ascontiguousarray(v, "d") for v in (low, high)]

    @staticmethod
    @njit(cache=True)
    def apply(minn, maxn, pop):
        """
        Update pop so all values lie within bounds
        """
        for y in pop:
            # Reflect points which are out of bounds
            idx = y < minn
            y[idx] = 2 * minn[idx] - y[idx]
            idx = y > maxn
            y[idx] = 2 * maxn[idx] - y[idx]

            # Randomize points which are still out of bounds
            idx = (y < minn) | (y > maxn)
            y[idx] = minn[idx] + util.rng.rand(sum(idx)) * (maxn[idx] - minn[idx])


class ClipBounds(Bounds):
    """
    Clip values to bounded region
    """

    c_interface = dll.bounds_clip if dll else None

    def __init__(self, low, high):
        self.low, self.high = [np.ascontiguousarray(v, "d") for v in (low, high)]

    @staticmethod
    @njit(cache=True)
    def apply(minn, maxn, pop):
        for y in pop:
            idx = y < minn
            y[idx] = minn[idx]
            idx = y > maxn
            y[idx] = maxn[idx]


class FoldBounds(Bounds):
    """
    Wrap values into the bounded region
    """

    c_interface = dll.bounds_fold if dll else None

    def __init__(self, low, high):
        self.low, self.high = [np.ascontiguousarray(v, "d") for v in (low, high)]

    @staticmethod
    @njit(cache=True)
    def apply(minn, maxn, pop):
        for y in pop:
            # Deal with semi-infinite cases using reflection
            idx = (y < minn) & isinf(maxn)
            y[idx] = 2 * minn[idx] - y[idx]
            idx = (y > maxn) & isinf(minn)
            y[idx] = 2 * maxn[idx] - y[idx]

            # Wrap points which are out of bounds
            idx = y < minn
            y[idx] = maxn[idx] - (minn[idx] - y[idx])
            idx = y > maxn
            y[idx] = minn[idx] + (y[idx] - maxn[idx])

            # Randomize points which are still out of bounds
            idx = (y < minn) | (y > maxn)
            y[idx] = minn[idx] + util.rng.rand(sum(idx)) * (maxn[idx] - minn[idx])


class RandomBounds(Bounds):
    """
    Randomize values into the bounded region
    """

    c_interface = dll.bounds_random if dll else None

    def __init__(self, low, high):
        self.low, self.high = [np.ascontiguousarray(v, "d") for v in (low, high)]

    @staticmethod
    @njit(cache=True)
    def apply(minn, maxn, pop):
        for y in pop:
            # Deal with semi-infinite cases using reflection
            idx = (y < minn) & isinf(maxn)
            y[idx] = 2 * minn[idx] - y[idx]
            idx = (y > maxn) & isinf(minn)
            y[idx] = 2 * maxn[idx] - y[idx]

            # The remainder are selected uniformly from the bounded region
            idx = (y < minn) | (y > maxn)
            y[idx] = minn[idx] + util.rng.rand(sum(idx)) * (maxn[idx] - minn[idx])


class IgnoreBounds(Bounds):
    """
    Leave values outside the bounded region
    """

    c_interface = dll.bounds_ignore if dll else None

    def __init__(self, low=None, high=None):
        self.low, self.high = [np.ascontiguousarray(v, "d") for v in (low, high)]

    @staticmethod
    def apply(minn, maxn, pop):
        pass


def test():
    """bounds handlers test"""
    from numpy.linalg import norm
    from numpy import array

    bounds = list(zip([5, 10], [-inf, -10], [-5, inf], [-inf, inf]))
    v = np.ascontiguousarray([[6, -12, 6, -12]], "d")
    for t in "none", "reflect", "clip", "fold", "randomize":
        w = make_bounds_handler(bounds, t)
        assert norm(w(v + 0) - v) == 0
    v = np.ascontiguousarray([[12, 12, -12, -12]], "d")
    for t in "none", "reflect", "clip", "fold":
        w = make_bounds_handler(bounds, t)
        assert norm(w(v.repeat(3, axis=0)) - w(v + 0)) == 0
    assert norm(make_bounds_handler(bounds, "none")(v + 0) - v) == 0
    assert norm(make_bounds_handler(bounds, "reflect")(v + 0) - [8, -32, 2, -12]) == 0
    assert norm(make_bounds_handler(bounds, "clip")(v + 0) - [10, -10, -5, -12]) == 0
    assert norm(make_bounds_handler(bounds, "fold")(v + 0) - [7, -32, 2, -12]) == 0
    w = make_bounds_handler(bounds, "randomize")(v + 0)
    assert 5 <= w[0, 0] <= 10 and w[0, 1] == -32 and w[0, 2] == 2 and w[0, 3] == -12
    v = np.ascontiguousarray([[20, 1, 1, 1]], "d")
    w = make_bounds_handler(bounds, "reflect")(v + 0)
    assert 5 <= w[0, 0] <= 10
    v = np.ascontiguousarray([[20, 1, 1, 1]], "d")
    w = make_bounds_handler(bounds, "fold")(v + 0)
    assert 5 <= w[0, 0] <= 10


if __name__ == "__main__":
    test()



"""
Fit job definition for the distributed job queue.
"""

import os
import sys
import json

try:
    import dill as pickle
except ImportError:
    import pickle

from . import cli
from . import __version__

# Site configuration determines what kind of mapper to use
# This should be true in cli.py as well
from .mapper import MPMapper as Mapper
from . import monitor
from .fitters import FitDriver


def fitservice(request):
    import matplotlib

    matplotlib.use("Agg")

    path = os.getcwd()

    service_version = __version__
    request_version = str(request["version"])
    if service_version != request_version:
        raise ValueError("fitter version %s does not match request %s" % (service_version, request_version))

    data = request["data"]
    model = str(data["package"])

    service_model_version = __version__
    request_model_version = str(data["version"])
    if service_model_version != request_model_version:
        raise ValueError(
            "%s version %s does not match request %s" % (model, service_model_version, request_model_version)
        )
    options = pickle.loads(str(data["options"]))
    problem = pickle.loads(str(data["problem"]))
    problem.store = path
    problem.output_path = os.path.join(path, "model")

    fitdriver = FitDriver(options.fit, problem=problem, **options)

    fitdriver.mapper = Mapper.start_mapper(problem, options.args)
    problem.show()
    print("#", " ".join(sys.argv))
    best, fbest = fitdriver.fit()
    cli.save_best(fitdriver, problem, best)
    matplotlib.pyplot.show()
    return list(best), fbest


class ServiceMonitor(monitor.TimedUpdate):
    """
    Display fit progress on the console
    """

    def __init__(self, problem, path, progress=60, improvement=60):
        monitor.TimedUpdate.__init__(self, progress=progress, improvement=improvement)
        self.path = path
        self.problem = problem
        self.images = []

    def show_progress(self, history):
        p = self.problem.getp()
        try:
            self.problem.setp(history.point[0])
            dof = self.problem.dof
            summary = self.problem.summarize()
        finally:
            self.problem.setp(p)

        status = {
            "step": history.step[0],
            "cost": history.value[0] / dof,
            "pars": history.point[0],
        }
        json_status = json.dumps(status)
        open(os.path.join(self.path, "status.json"), "wt").write(json_status)
        status["table"] = summary
        status["images"] = "\n".join('<img file="%s" alt="%s" />' % (f, f) for f in self.images)
        html_status = (
            """\
<html><body>
Generation %(step)d, chisq %(cost)g
<pre>
%(table)s
</pre>
%(images)s
</body></html>
"""
            % status
        )
        open(os.path.join(self.path, "status.html"), "wt").write(html_status)

    def show_improvement(self, history):
        import pylab

        # print "step",history.step[0],"chisq",history.value[0]
        self.problem.setp(history.point[0])
        pylab.cla()
        self.problem.plot(figfile=os.path.join(self.path, "K"))
        pylab.gcf().canvas.draw()



import sys

import wx


class FitView(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent=parent, id=wx.ID_ANY)
        self.parent = parent

        sizer1 = wx.BoxSizer(wx.HORIZONTAL)
        label1 = wx.StaticText(self, 1, label="Store Folder:")

        self.store_file = wx.TextCtrl(self, 2, value="", style=wx.TE_RIGHT)
        sizer1.Add(label1, 0, border=5, flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL | wx.ALL)
        sizer1.Add(self.store_file, 1, wx.EXPAND | wx.RIGHT, border=10)

        # Create the Fit button.
        self.btn_fit = wx.Button(self, wx.ID_ANY, "Fit")
        self.btn_fit.SetToolTip(wx.ToolTip("click to start fit"))
        self.Bind(wx.EVT_BUTTON, self.OnFit, self.btn_fit)

        # Create a horizontal box sizer for the buttons.
        sizer2 = wx.BoxSizer(wx.HORIZONTAL)
        sizer2.Add(self.btn_fit, 0, wx.ALL, 5)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add((10, 10), 0)  # whitespace
        sizer.Add(sizer1, 0, wx.ALL, 5)
        sizer.Add(sizer2, 0, wx.ALL, 5)

        self.SetSizer(sizer)
        self.SetAutoLayout(True)

    def OnFit(self, event):
        btnLabel = self.btn_fit.GetLabel()
        if btnLabel == "Fit":
            self.btn_fit.SetLabel("Stop")
            self.store = self.store_file.GetValue()
            ################LOGIC######################
            # send fit event message to panel with
            # all required data to fit
            # the panel will listen to event and start
            # the fit.
            ###########################################
            send("fit", store=self.store)

        else:
            print("stop logic goes here")
            self.btn_fit.SetLabel("Fit")
            pass

    def OnFitComplete(self, event):
        self.btn_fit.SetLabel("Fit")
